"""Module that supports learning to do auto transactions."""


def get_bookings(booking, line):
    """Take line and try to deduce if it is auto bookable.

    Try to deduce from past experience if the booking identifies to second
    leg if it does returns array of bookings than are directly bookable.

    Otherwise returns empty list(array)
    """
    legs = []
    partner = booking["partner"]
    auto_booking = False
    if partner == "SEB banka":
        auto_booking = "seb-commission"
    elif "pamatsumma" in line[9] and "KT08093" in line[9]:
        auto_booking = "private-expense"
    elif "Paysafe Payment Solutions" == partner or "Linda Vernava" in partner:
        auto_booking = "private-income"
    if auto_booking:
        booking = dict(booking)
        booking["book_type"] = auto_booking
        legs.append(booking)
    return legs



"""Tallybot plugin to zoozl server."""

import time

from agents import set_default_openai_key, Runner, SQLiteSession
from zoozl import agentgear, utils
from zoozl.chatbot import Interface, Package, InterfaceRoot

from . import agents, brain



class Handler(agentgear.StreamHandler):
    """Handler of the single stream."""

    def handle_event(self, event):
        """Handle event."""
        print("Event:", event.event)

    def on_requires_action(self, action):
        """Handle action."""
        print("Action:", action)
        self.client.beta.threads.runs.submit_tool_outputs(
            self.current_run.id,
            thread_id=self.current_run.thread_id,
            tool_outputs=[
                {
                    "tool_call_id": tool_call.id,
                    "output": output,
                }
            ],
        )


    def on_tool_call_created(self, tool_call):
        """Handle tool call."""
        if tool_call.type == "function":
            if tool_call.function.name.startswith("transfer_to_"):
                assistant_id = self.assistant_map[tool_call.function.name[12:]]
                thread = self.client.beta.threads.create()
                self.client.beta.threads.messages.create(
                    thread.id,
                    role="user",
                    content=self.package.last_message.text,
                )
                make_a_run(
                    self.client,
                    thread.id,
                    assistant_id,
                    self.assistant_map,
                    self.package,
                    self.context
                )
                self.package.conversation.data["openai_thread_id"] = thread.id
                self.package.conversation.data["openai_assistant_id"] = assistant_id
                output = "Transfer complete"
            else:
                msg, attach, fname = brain.do_task(
                    self.context.conf["tallybot"],
                    self.context.memory,
                    tool_call.function.name,
                    tool_call.function.arguments,
                    None
                )
                output = msg

def make_a_run(
    assistant_id: str,
    assistant_map: dict,
    package: Package,
    context: InterfaceRoot
) -> None:
    """Make a single run."""
    with client.beta.threads.runs.stream(
        thread_id=thread_id,
        assistant_id=assistant_id,
        timeout=10,
        event_handler=Handler(client, thread_id, assistant_map, package, context),
    ) as stream:
        stream.until_done()
        run = stream.get_final_run()
        while run.completed_at is None:
            if run.created_at < int(time.time()) - 10:
                stream.on_timeout()
                run = client.beta.threads.runs.cancel(run.id, thread_id=thread_id)
                checks = 0
                while run.cancelled_at is None and checks < 10:
                    time.sleep(3)
                    run = client.beta.threads.runs.poll(run.id, thread_id=thread_id)
                    checks += 1
                break
            else:
                time.sleep(3)
                run = client.beta.threads.runs.poll(run.id, thread_id=thread_id)


class TallyBot(Interface):
    """Tallybot interface to zoozl chatbot."""

    aliases = {"tallybot", "help", "greet", "cancel"}

    def load(self, root: InterfaceRoot):
        """Load OpenAI agents."""
        set_default_openai_key(root.conf["tallybot"]["openai_api_key"])
        try:
            api_key = root.conf["tallybot"]["openai_api_key"]
        except KeyError:
            raise RuntimeError(
                "Tallybot requires openAI api key to work!"
            ) from None
        self.assistant_map = {"Tallybot": agents.tallybot}
        self.db_path = root.conf["tallybot"]["database"]

    def consume(self, package: Package):
        """Handle incoming message."""
        session = SQLiteSession(package.talker, self.db_path)
        run = Runner.run_sync(self.assistant_map["Tallybot"], package.last_message.text, session=session)
        package.callback(run.final_output)



"""Module defines chat subject interfaces."""
import dataclasses as data
from dataclasses import dataclass

from . import process


@dataclass
class Interface():
    """Interface to the chat command handling."""

    required: set = data.field(default_factory=set)
    optional: set = data.field(default_factory=set)
    attachment: bool = False
    # must contain callables that accept text and conversation memory item, if non empty
    auxiliaries: set = data.field(default_factory=set)
    one_optional: bool = False  # if True, at least one optional field is required


def do_add_default_partner():
    """Add default partner by it's name."""
    return Interface(
        required={"name"},
    )


def do_update_myself():
    """Update myself data."""
    return Interface(
        optional={
            "name",
            "address1",
            "address2",
            "address3",
            "address4",
            "phone",
            "email",
            "bank_name",
            "bank_account",
            "bank_code",
            "rate",
            "currency"},
        one_optional=True,
    )


def do_create_invoice():
    """Create a new invoice."""
    return Interface(
        required={"week1"},
        optional={"week" + str(i) for i in range(2, 6)},
    )


def do_book_invoice():
    """Book attached invoice."""
    return Interface(attachment=True)


def do_private_income():
    """Do private income."""
    return Interface(
        required={"date", "partner", "value"},
    )


def do_get_partner():
    """Get partner data."""
    return Interface(
        required={"name"},
    )


def do_remove_transaction():
    """Remove transaction."""
    return Interface(
        required={"id"},
    )


def do_transaction():
    """Create new transaction."""
    return Interface(
        required={
            "date",
            "reference",
            "source",
            "comment",
            "debit_amount",
            "debit",
            "credit",
        },
        optional={
            "partner",
            "credit_amount",
            "deal_value",
            "rate",
            "debit_currency",
            "credit_currency",
            "debit_stack",
            "credit_stack",
        },
    )


def do_add_expense():
    """Define add expense subject requirements."""
    return Interface(
        required={"date", "reference", "comment", "partner", "value"},
        optional={"expense_account", "currency", "split"},
        attachment=True,
        auxiliaries={allow_no_attachment, }
    )


def do_private_expense():
    """Private expense."""
    return Interface(
        required={"date", "partner", "value"},
    )


def do_get_help():
    """Define help requirements."""
    return Interface()


def do_get_myself():
    """Define myself requirements."""
    return Interface()


def do_recalculate_outstanding():
    """Recalculate outstanding items per partner."""
    return Interface(
        required={"year", "partner"},
    )


def do_get_outstanding():
    """Define report requirements."""
    return Interface()


def do_create_partner():
    """Define partner requirements."""
    return Interface(
        required={"name"},
        optional={"other_names", "id"},
    )


def do_update_partner():
    """Define partner update requirements."""
    return Interface(
        required={"name"},
        optional={
            "reg_no",
            "bank_account",
            "bank_name",
            "invoice_prefix",
            "invoice_row_format",
            "terms",
            "currency",
            "address1",
            "address2",
            "address3",
            "address4",
            "new_name",
        }
    )


def allow_no_attachment(text, talk):
    """Support function to drop attachment requirement."""
    cmds = ["no attachment", "without attachment", "skip attachment", "drop attachment"]
    pos = process.extractOne(text, cmds)
    if pos[1] > 98:
        talk.attachment = b'Attachment explicitly skipped'
        return "Skipping attachment"
    return ""


def do_upwork_invoices():
    """Upwork invoice upload."""
    return Interface(
        attachment=True,
    )


def do_seb_statement():
    """Seb statement upload."""
    return Interface(
        attachment=True,
    )


def do_upwork_statement():
    """Upwork statement upload."""
    return Interface(
        attachment=True,
    )


def do_get_ledger():
    """Full ledger report in excel."""
    return Interface(
        required={"filter_by_quarter"},
    )


def do_update_ledger():
    """Manual ledger update."""
    return Interface(
        attachment=True,
    )



"""Loads all tests automatically from tests module.

Recurses into tests directory and assumes all .py extensions to be test
files, except ones that start with _
"""

import os


def load_tests(loader, standard_tests, pattern):
    """Unittest calls function to load tests from package."""
    pattern = "[!_]*.py"
    this_dir = os.path.dirname(__file__)
    package_tests = loader.discover(start_dir=this_dir, pattern=pattern)
    standard_tests.addTests(package_tests)
    return standard_tests



"""Invoice layout."""
import enum
import datetime
import decimal as d

from reportlab import platypus as p
from reportlab.rl_config import defaultPageSize
from reportlab.lib import colors
from reportlab.lib.styles import StyleSheet1, ParagraphStyle
from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont

from . import schema


pdfmetrics.registerFont(
    TTFont("Font", "/usr/share/fonts/truetype/dejavu/DejaVuSansMono.ttf"))
pdfmetrics.registerFont(
    TTFont("Font-Bold", "/usr/share/fonts/truetype/dejavu/DejaVuSansMono-Bold.ttf"))

c = d.getcontext()
c.traps[d.FloatOperation] = True

PAGE_HEIGHT = defaultPageSize[1]
PAGE_WIDTH = defaultPageSize[0]

# Define styles used in invoice document
DEFAULT_FONT = "Font"
DEFAULT_FONT_BOLD = "Font-Bold"
COLOR_GREY = colors.Color(0.9, 0.9, 0.9)


def get_style_book():
    """Prepare all styles used for invoice document."""
    style_book = StyleSheet1()
    style_book.add(ParagraphStyle(
        "default",
        fontName=DEFAULT_FONT,
    ))
    style_book.add(ParagraphStyle(
        "default-bold",
        fontName=DEFAULT_FONT_BOLD,
    ))
    style_book.add(ParagraphStyle(
        "HeaderLeft",
        style_book["default"],
        alignment=TA_LEFT,
    ))
    style_book.add(ParagraphStyle(
        "HeaderLeftBold",
        style_book["default-bold"],
        alignment=TA_LEFT,
    ))
    style_book.add(ParagraphStyle(
        "HeaderRight",
        style_book["default"],
        alignment=TA_RIGHT,
    ))
    style_book.add(ParagraphStyle(
        "HeaderRightBold",
        style_book["default-bold"],
        alignment=TA_RIGHT,
    ))
    style_book.add(ParagraphStyle(
        "HeaderTitle",
        style_book["default-bold"],
        alignment=TA_CENTER,
        fontSize=12,
    ))
    style_book.add(ParagraphStyle(
        "TableHeader",
        style_book["default-bold"],
    ))
    style_book.add(ParagraphStyle(
        "TableHeaderCenter",
        style_book["TableHeader"],
        alignment=TA_CENTER,
    ))
    style_book.add(ParagraphStyle(
        "Table",
        style_book["default"],
    ))
    style_book.add(ParagraphStyle(
        "TableCenter",
        style_book["Table"],
        alignment=TA_CENTER,
    ))
    style_book.add(ParagraphStyle(
        "Footer",
        style_book["default"],
        fontSize=6,
    ))
    return style_book


STYLES = get_style_book()


class FirstPage(int, enum.Enum):
    """First page of the invoice.

    First page contains 4 frames:
    - Biller info & Payer info
    - Invoice header
    - Invoice table
    - Invoice payment info

    +----------+      +-----------+
    |          |      |           |
    |  BILLER  |      |  HEADER   |
    |          |      |           |
    +----------+      +-----------+

    +-----------------------------+
    |                             |
    |           TABLE             |
    |                             |
    +-----------------------------+

    +---------------+
    |     PAYTO     |
    +---------------+

    +----------FOOTER-------------+
    """

    HEIGHT = defaultPageSize[1]
    WIDTH = defaultPageSize[0]
    TOP_MARGIN = 60
    LEFT_MARGIN = 30
    RIGHT_MARGIN = 30
    BOTTOM_MARGIN = 5


class Biller(int, enum.Enum):
    """Positions of payer and biller frame.

    +----------+   +-----------+
    |  BILLER  |   |           |
    +----------+   |  DETAILS  |
                   |           |
                   |           |
                   +-----------+
    +----------+   +-----------+
    |  PAYER   |   |           |
    +----------+   |  DETAILS  |
                   |           |
                   |           |
                   +-----------+
    """

    TOP_MARGIN = FirstPage.TOP_MARGIN + 20
    LEFT_MARGIN = FirstPage.LEFT_MARGIN + 0
    DETAILS_HEIGHT = 100
    DETAILS_WIDTH = 170
    TITLE_HEIGHT = 30
    TITLE_WIDTH = 55
    SPACE = 10


def get_biller_frame():
    """Return biller frame."""
    return frame_left_corner(
        Biller.LEFT_MARGIN,
        Biller.TOP_MARGIN,
        Biller.TITLE_WIDTH + Biller.DETAILS_WIDTH,
        2 * Biller.DETAILS_HEIGHT + Biller.SPACE
    )


def create_biller_info(biller_text, payer_text):
    """Create a biller and payer info.

    Args:
        biller_text: iterator of strings
        payer_text: iterator of strings

    Return list of shapes(flowables) that should be added in biller frame.
    """
    b_head = p.Paragraph("From:", STYLES["HeaderLeftBold"])
    p_head = p.Paragraph("Bill to:", STYLES["HeaderLeftBold"])
    biller = p.Paragraph("<br/>".join(biller_text), STYLES["default"])
    payer = p.Paragraph("<br/>".join(payer_text), STYLES["default"])
    return [p.Table(
        [[b_head, biller], [p.Spacer(1, 0), p.Spacer(1, 0)], [p_head, payer]],
        colWidths=[Biller.TITLE_WIDTH, Biller.DETAILS_WIDTH],
        rowHeights=[
            Biller.DETAILS_HEIGHT, Biller.SPACE, Biller.DETAILS_HEIGHT],
        style=[
            ("VALIGN", (0, 0), (-1, -1), "TOP"),
            ("LEFTPADDING", (0, 0), (-1, -1), 0),
            ("RIGHTPADDING", (0, 0), (-1, -1), 0),
        ]
    )]


class Header(int, enum.Enum):
    """Invoice header frame positions.

    Invoice header is divided into 3 columns:
     +-----------------------------+
     |           TITLE             |
     +-----------------------------+

      +-----------+ +-------------+
      |           | |             |
      | LEFT_COL  | |  RIGHT_COL  |
      |           | |             |
      |           | |             |
      +-----------+ +-------------+
    """

    TOP_MARGIN = FirstPage.TOP_MARGIN + 30
    RIGHT_MARGIN = FirstPage.RIGHT_MARGIN + 0
    HEIGHT = 100
    WIDTH = 200
    TITLE_HEIGHT = 30
    SPACE_AFTER_TITLE = 10
    RIGHT_COL_WIDTH = 100
    MARGIN_COL = 10


def get_invoice_header_frame():
    """Return invoice header frame."""
    return frame_right_corner(
        Header.RIGHT_MARGIN,
        Header.TOP_MARGIN,
        Header.WIDTH,
        Header.HEIGHT)


def create_invoice_heading(i: schema.InvoiceData):
    """Create an invoice heading.

    Args:
       values: list of strings

    Invoice header contains information:
    - Invoice number
    - Invoice date
    - Invoice due date
    - Invoice total amount
    - Invoice total due

    Return a list of shapes(flowables) that should be added in header frame.
    """
    date = datetime.date.today()
    due_date = date + datetime.timedelta(days=i.terms)
    values = [
        i.no,
        date.strftime("%d %b %Y"),
        due_date.strftime("%d %b %Y"),
        i.currency + ' ' + i.total,
        i.currency + ' ' + i.total]
    shapes = []
    title = p.Paragraph("I N V O I C E", STYLES["HeaderTitle"])
    bold_ind = [2, 4]
    headings = [
        "INVOICE #", "INVOICE DATE", "DUE DATE", "TOTAL AMOUNT", "TOTAL DUE"]
    left_col = apply_style_to_list(
        headings, STYLES["HeaderLeft"], STYLES["HeaderLeftBold"], bold_ind)
    right_col = apply_style_to_list(
        values, STYLES["HeaderRight"], STYLES["HeaderRightBold"], bold_ind)
    shapes.append(p.Table(
        [[title]],
        colWidths=[Header.WIDTH],
        rowHeights=[Header.TITLE_HEIGHT],
        style=[
            ("BACKGROUND", (0, 0), (-1, -1), COLOR_GREY),
            ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
        ]
    ))
    shapes.append(p.Spacer(0, Header.SPACE_AFTER_TITLE))
    row_t_height =\
        Header.HEIGHT\
        - Header.TITLE_HEIGHT\
        - Header.SPACE_AFTER_TITLE
    shapes.append(p.Table(
        list(zip(left_col, right_col)),
        colWidths=[
            Header.WIDTH - Header.RIGHT_COL_WIDTH - 2*Header.MARGIN_COL,
            Header.RIGHT_COL_WIDTH
        ],
        rowHeights=row_t_height/len(left_col),
        style=[
            ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
            ("ALIGN", (0, 0), (0, -1), "LEFT"),
            ("ALIGN", (1, 0), (1, -1), "RIGHT"),
            ("LEFTPADDING", (0, 0), (-1, -1), 0),
            ("RIGHTPADDING", (0, 0), (-1, -1), 0),
        ]
    ))
    return shapes


class Table(int, enum.Enum):
    """Invoice table frame positions.

    +-------------------------------+
    | MEMO    | HOURS | RATE | TOTAL|
    +-------------------------------+
    |         |       |      |      |
    |         |       |      |      |
    |         |       |      |      |
    +-------------------------------+
    """

    TOP_MARGIN = Biller.TOP_MARGIN + 2 * Biller.DETAILS_HEIGHT + Biller.SPACE
    SPACE_BEFORE = 70
    LEFT_MARGIN = FirstPage.LEFT_MARGIN + 0
    HEIGHT = 300
    WIDTH = FirstPage.WIDTH - FirstPage.LEFT_MARGIN - FirstPage.RIGHT_MARGIN
    HOURS_WIDTH = 50
    RATE_WIDTH = 50
    TOTAL_WIDTH = 65


def get_table_frame():
    """Return invoice table frame."""
    return frame_left_corner(
        Table.LEFT_MARGIN,
        Table.TOP_MARGIN + Table.SPACE_BEFORE,
        Table.WIDTH,
        Table.HEIGHT)


def create_table(rows, rate, total):
    """Create an invoice table."""
    style = [
        ('GRID', (0, 0), (-1, -1), 0.1, colors.black),
        ("BACKGROUND", (0, 0), (-1, 0), COLOR_GREY),
        ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
    ]
    cells = [[
        p.Paragraph("DESCRIPTION", STYLES["TableHeader"]),
        p.Paragraph("HOURS", STYLES["TableHeaderCenter"]),
        p.Paragraph("RATE", STYLES["TableHeaderCenter"]),
        p.Paragraph("TOTAL", STYLES["TableHeaderCenter"]),
    ]]
    for item in rows:
        cells.append([
            p.Paragraph(item.description, STYLES["Table"]),
            p.Paragraph(str(item.hours), STYLES["TableCenter"]),
            p.Paragraph(str(rate), STYLES["TableCenter"]),
            p.Paragraph(str(item.total), STYLES["TableCenter"]),
        ])
    cells.append([
        p.Paragraph("TOTAL", STYLES["Table"]),
        p.Paragraph("", STYLES["TableCenter"]),
        p.Paragraph("", STYLES["TableCenter"]),
        p.Paragraph(str(total), STYLES["TableCenter"]),
    ])
    return [p.Table(
        cells,
        style=style,
        colWidths=[
            Table.WIDTH
            - Table.HOURS_WIDTH
            - Table.RATE_WIDTH
            - Table.TOTAL_WIDTH,
            Table.HOURS_WIDTH,
            Table.RATE_WIDTH,
            Table.TOTAL_WIDTH],
        rowHeights=25)]


class PayInfo(int, enum.Enum):
    """Invoice pay info frame positions.

    +--------------------------------+
    |           PAY TO               |
    +--------+-----------------------+
    |        |                       |
    |        |                       |
    | LABEL  |  INFORMATION          |
    |        |                       |
    |        |                       |
    |        |                       |
    +--------+-----------------------+
    """

    BOTTOM_MARGIN = 70
    LEFT_MARGIN = FirstPage.LEFT_MARGIN + 0
    TITLE_HEIGHT = 40
    ROW_HEIGHT = 20
    WIDTH = 250
    LABEL_WIDTH = 40


def get_pay_info_frame():
    """Return invoice pay info frame."""
    return blank_frame(
        PayInfo.LEFT_MARGIN,
        PayInfo.BOTTOM_MARGIN,
        PayInfo.WIDTH,
        3*PayInfo.ROW_HEIGHT + PayInfo.TITLE_HEIGHT)


def create_pay_info(bank: schema.BankInfo):
    """Create an invoice pay info frame."""
    style = [
        ("LEFTPADDING", (0, 0), (-1, -1), 2),
        ("RIGHTPADDING", (0, 0), (-1, -1), 0),
    ]
    cells = [[
        p.Paragraph("Bank:", STYLES["Table"]),
        p.Paragraph(bank.name, STYLES["Table"]),
    ]]
    cells.append([
        p.Paragraph("IBAN:", STYLES["Table"]),
        p.Paragraph(bank.iban, STYLES["Table"]),
    ])
    cells.append([
        p.Paragraph("BIC:", STYLES["Table"]),
        p.Paragraph(bank.bic, STYLES["Table"]),
    ])
    return [
        p.Table(
            [[p.Paragraph("PAY BY BANK TRANSFER", STYLES["TableHeader"])]],
            style=[
                ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ("BACKGROUND", (0, 0), (-1, 0), COLOR_GREY)],
            colWidths=PayInfo.WIDTH,
            rowHeights=PayInfo.TITLE_HEIGHT),
        p.Table(
            cells,
            style=style,
            colWidths=[
                PayInfo.LABEL_WIDTH,
                PayInfo.WIDTH - PayInfo.LABEL_WIDTH],
            rowHeights=PayInfo.ROW_HEIGHT)
    ]


def get_footer_frame():
    """Return invoice footer frame."""
    return blank_frame(
        FirstPage.LEFT_MARGIN,
        FirstPage.BOTTOM_MARGIN,
        FirstPage.WIDTH - FirstPage.LEFT_MARGIN - FirstPage.RIGHT_MARGIN,
        15)


def create_footer():
    """Add footer with thanks to."""
    msg = "*This invoice was generated automatically using "
    msg += "open source software (Python, ReportLab)"
    return [p.Paragraph(
        msg,
        STYLES["Footer"],
    )]


def apply_style_to_list(lst, n_style, b_style, b_rows):
    """Apply style.

    Args:
        lst: list of strings
        n_style: normal Paragraph style
        b_style: bold style
        b_rows: indexes to apply bold style
        kwargs: additional arguments to ParagraphStyle
    """
    applied_style = []
    for i, item in enumerate(lst):
        if i in b_rows:
            applied_style.append(p.Paragraph(item, style=b_style))
        else:
            applied_style.append(p.Paragraph(item, style=n_style))
    return applied_style


def blank_frame(*args, **kwargs):
    """Create a blank frame.

    Remove default padding from the frame.
    """
    for pad in ["leftPadding", "rightPadding", "topPadding", "bottomPadding"]:
        if pad not in kwargs:
            kwargs[pad] = 0
    return p.Frame(*args, **kwargs)


def frame_left_corner(x, y, width, height):
    """Create a frame in top left corner.

    Args:
        x: margin from the left
        y: margin from the top
        width: width of the frame
        height: height of the frame
    """
    return blank_frame(x, PAGE_HEIGHT - y - height, width, height)


def frame_right_corner(x, y, width, height):
    """Create a frame in top right corner.

    Args:
        x: margin from the right
        y: margin from the top
        width: width of the frame
        height: height of the frame
    """
    return blank_frame(
        PAGE_WIDTH - x - width,
        PAGE_HEIGHT - y - height,
        width,
        height)


def get_invoice_template(bytes_io):
    """Generate an invoice document template with empty frames.

    Args:
        bytes_io: BytesIO object to write the binary document to
    """
    template = p.PageTemplate(
        frames=[
            get_biller_frame(),
            get_invoice_header_frame(),
            get_table_frame(),
            get_pay_info_frame(),
            get_footer_frame(),],
    )
    doc = p.BaseDocTemplate(
        bytes_io, pageTemplates=[template])
    return doc


def strnum(num):
    """Return a number with leading zeros."""
    return f"{num:,.2f}"


def write_invoice(bytes_io, inv: schema.InvoiceData):
    """Write invoice to a file.

    Args:
        bytes_io: BytesIO object to write the binary document to
    """
    total = d.Decimal(0)
    for item in inv.rows:
        total += d.Decimal(item.hours) * d.Decimal(inv.rate)
        item.total = strnum(d.Decimal(item.hours) * d.Decimal(inv.rate))
    inv.total = strnum(total)
    elements = []
    for item in [
            create_biller_info(inv.biller, inv.payer),
            create_invoice_heading(inv),
            create_table(
                inv.rows,
                rate=strnum(d.Decimal(inv.rate)),
                total=strnum(total)),
            create_pay_info(inv.biller_bank),
            create_footer()]:
        elements += item
        elements.append(p.FrameBreak())
    doc = get_invoice_template(bytes_io)
    doc.author = "Juris Kaminskis"
    doc.creator = "Juris Kaminskis"
    doc.title = "Invoice"
    doc.subject = "Invoice"
    doc.build(elements)


if __name__ == "__main__":
    """Create an invoice document 'invoice.pdf' in the current directory."""
    from io import BytesIO
    data = schema.InvoiceData(
        "EXN987",
        [
            "ICO Reklamation",
            "Forest street 21",
            "Alembotan",
        ],
        schema.BankInfo("Bank name", "Bank account", "Bank code"),
        ["John Doe", "New York", " USA+1 123456789"],
        (
            schema.InvoiceRow("Some meaningfull text on the invoice", "21"),
            schema.InvoiceRow("Other services", "20")),
        40,
        "50.1",
        "EUR"
    )
    # Write pdf out
    writer = BytesIO()
    write_invoice(writer, data)
    with open("invoice.pdf", "wb") as f:
        f.write(writer.getvalue())



"""Network related requests."""

import socket
import time


def wait_for_connection(port, timeout=2):
    """Open socket and wait for connection."""
    start = time.time()
    while start + timeout > time.time():
        sock = socket.socket()
        try:
            sock.connect(("localhost", port))
        except ConnectionRefusedError:
            time.sleep(0.01)
        else:
            sock.shutdown(socket.SHUT_RDWR)
            return
        finally:
            sock.close()
    raise AssertionError("Did not establish connection within timeout")



"""Chat type interaction for accountant."""

import membank

from . import brain, memories, chat_interface, handlers, process


def conversation(talker, text, files, callback, conf):
    """Make conversation.

    Receive text and send message back via callback.
    """
    talk = Chat(talker, conf, callback)
    if talk.ongoing:
        if talk.subject:
            talk.do_subject(text, files)
            talk.action()
        else:
            if not talk.get_subject(text):
                callback(
                    "I didn't get. Would you like me to send full list of commands?"
                )
                talk.set_subject("do_get_help")
            else:
                talk.action()
    else:
        talk.ongoing = True
        if not talk.get_subject(text):
            callback("What would you like me to do?")
        else:
            talk.action()


# pylint: disable=too-many-instance-attributes
class Chat:
    """Interface for communication and routing with talker."""

    def __init__(self, talker, conf, callback):
        """Initialise comm interface with talker."""
        self._m = membank.LoadMemory(conf.get("db_path"))
        self._t = self._m.get.conversation(talker=talker)
        if not self._t:
            self._t = memories.Conversation(
                talker=talker,
            )
        self._call = callback
        self._conf = conf
        self._interface = chat_interface.Interface()
        self._interface_ready = self._load_interface()

    @property
    def talker(self):
        """Return talker."""
        return self._t.talker

    @property
    def ongoing(self):
        """Check if talk is ongoing."""
        return self._t.ongoing

    @ongoing.setter
    def ongoing(self, value):
        """Set talk ongoing value."""
        self._t.ongoing = value
        self._m.put(self._t)

    @property
    def subject(self):
        """Return subject if present."""
        return self._t.subject

    def action(self):
        """Check if ready for action, if so executes otherwise asks info."""
        if self.is_complete():
            self.perform()
        else:
            self.ask_missing()

    def get_subject(self, text):
        """Try to understand subject from text.

        if understood set the subject and return it otherwise return
        None.
        """
        subj = self._m.get.chatsubjectlookup(sentence=text)
        if subj:
            self.set_subject(subj.cmd)
        else:
            cmd_list = [
                " ".join(i.split("_")) for i in brain.generate_cmd_list()
            ]
            pos = process.extractOne(text, cmd_list)
            if pos[1] >= 95:
                cmd = pos[0].replace(" ", "_")
                self._m.put(memories.ChatSubjectLookup(cmd, text))
                subj = cmd
                self.set_subject(subj)
        return subj

    def set_subject(self, cmd):
        """Set subject as per cmd."""
        self._t.subject = cmd
        self._m.put(self._t)
        self._interface_ready = self._load_interface()

    def clear_subject(self):
        """Reset conversation to new start."""
        self.set_subject("")
        self._call("OK. Let's start over.")

    def do_subject(self, text, attach):
        """Continue on the subject."""
        if self._positive(text):
            self._load_data(text, attach[0])

    def _positive(self, text):
        """Assert that text seems positive otherwise cancels the subject."""
        choices = [
            "no",
            "cancel",
            "stop",
            "stop it",
            "forget",
            "start again",
            "naah",
        ]
        pos = process.extractOne(text, choices)
        if pos[1] > 97:
            self.clear_subject()
            return False
        return True

    def _load_data(self, text, attachment):
        """Load text and attachment as per subject interface."""
        auxes = False
        if self._interface.auxiliaries:
            for support in self._interface.auxiliaries:
                answer = support(text, self._t)
                if answer:
                    self._call(answer)
                    self._m.put(self._t)
                    auxes = True
                    break
        if auxes:
            return  # noqa: E701
        try:
            data = handlers.get_json(text)
        except RuntimeWarning as err:
            self._call(err)
        if attachment and self._interface.attachment:
            self._t.attachment = attachment
            self._m.put(self._t)
        elif attachment:
            self._call("I don't need attachment so I ignore it")
        fields = self._interface.required.union(self._interface.optional)
        if len(data) > 1:
            self._call("Sorry not more than record. Do not duplicate fields")
        elif len(data) == 1:
            data = data[0]
            for key in data:
                if key in fields:
                    self._t.data[key] = data[key]
                    self._m.put(self._t)
                else:
                    self._call(f"Field '{key}' is not valid, ignoring")

    def _load_interface(self):
        """Load interface demands marks internal flag if succesfull."""
        if not self._t.subject:
            return False
        if self._t.subject not in dir(chat_interface):
            msg = f"Ups. I am confused, don't know what to do with '{self._t.subject}'"
            self._call(msg)
            return False
        interface = getattr(chat_interface, self._t.subject)
        self._interface = interface()
        return True

    def is_complete(self):
        """Check if subject interface is complete and ready to pass on."""
        if not self._interface_ready:
            return False
        keys = self._t.data.keys()
        for req in self._interface.required:
            if req not in keys:
                return False
        if self._interface.one_optional and not keys:
            return False
        return self._interface.attachment == bool(self._t.attachment)

    def perform(self):
        """Perform a task."""
        msg, attach, fname = brain.do_task(
            self._conf,
            self._m,
            self._t.subject,
            [self._t.data],
            self._t.attachment,
        )
        self._call(msg, attach, fname)
        self._t.subject = ""
        self._t.attachment = b""
        self._t.data = {}
        self._m.put(self._t)

    def ask_missing(self):
        """Send missing data requirements."""
        msg = ""
        keys = self._t.data.keys()
        mand = []
        for req in self._interface.required:
            if req not in keys:
                mand.append(req)
        attach = self._interface.attachment == bool(self._t.attachment)
        opts = []
        for opt in self._interface.optional:
            if opt not in keys:
                opts.append(opt)
        if mand:
            msg += "I miss fields: " + ", ".join(mand)
        if opts:
            if msg:
                msg += "\n"
            msg += "Optional fields: " + ", ".join(opts)
        if not attach:
            if msg:
                msg += "\n"
            msg += "I need attachment"
        if msg:
            self._call(msg)



"""Package contains all functions accountant is able to logically do.

>>> do_task(conf, mem, "do_get_help", {}, b'')
"""
from .main import do_task, generate_cmd_list

__all__ = ["do_task", "generate_cmd_list"]



"""Functions related to creating bookings in ledger."""
import datetime

from ..memories import Transaction, ReportStruct, PARTNER_ACCOUNTS
from ..handlers import ExcelGenerator
from ..lookups import get_partner


def get_debitsum_by_partner(mem, year, partner, account):
    """Return sum of all transaction debit amounts for a partner."""
    return sum(
        i.debit_amount for i in get_partner_bookings(mem, year, partner, debit=account)
    )


def get_creditsum_by_partner(mem, year, partner, account):
    """Return sum of all transaction credit amounts for a partner."""
    return sum(
        i.credit_amount for i in get_partner_bookings(mem, year, partner, credit=account)
    )


def get_partner_bookings(mem, year, partner, debit=None, credit=None):
    """Return sum of all transaction for a partner."""
    this_year = datetime.date(year=year, month=1, day=1)
    next_year = datetime.date(year=year+1, month=1, day=1)
    partner = get_partner(mem, partner)
    filters = {"partner": partner}
    if debit:
        filters["debit"] = debit
    if credit:
        filters["credit"] = credit
    return mem.get(
        "transaction",
        mem.transaction.date >= this_year,
        mem.transaction.date < next_year,
        **filters
    )


def get_previous_quarter(today=None):
    """Return start date and end date for previous quarter.

    Depending on today's date or given date, return tuple of date objects
    in format of [start, end), where start is inclusive and end is exclusive.
    """
    if today is None:
        today = datetime.date.today()
    if today.month < 4:
        start = datetime.date(year=today.year-1, month=10, day=1)
        end = datetime.date(year=today.year, month=1, day=1)
    else:
        current_quarter = (today.month - 1) // 3 + 1
        q_month = (current_quarter - 2) * 3 + 1
        start = datetime.date(year=today.year, month=q_month, day=1)
        if start.month == 10:
            end = datetime.date(year=start.year+1, month=1, day=1)
        else:
            end = datetime.date(year=start.year, month=start.month+3, day=1)
    return start, end


def get_interval_dates(data):
    """From data dictionary input deduce start and end date for ledger.

    Expects dictionary returns tuple that is date interval of start and end
    (excluding)
    In case fails returns False in both values.
    """
    start = end = False
    if "filter_by_quarter" in data:
        if data["filter_by_quarter"].startswith("q"):
            q_month = int(data["filter_by_quarter"][1]) * 3 - 2
            today = datetime.date.today()
            start = datetime.date(year=today.year, month=q_month, day=1)
            if start.month == 10:
                end = datetime.date(year=start.year+1, month=1, day=1)
            else:
                end = datetime.date(year=start.year, month=start.month+3, day=1)
        elif data["filter_by_quarter"].startswith("l"):
            start, end = get_previous_quarter()
    elif "filter_by_month" in data:
        start = datetime.datetime.strptime(data["filter_by_month"], "%Y-%m")
        start = datetime.date(year=start.year, month=start.month, day=1)
        if start.month == 12:
            end = datetime.date(year=start.year+1, month=1, day=1)
        else:
            end = datetime.date(year=start.year, month=start.month+1, day=1)
    return start, end


def add_ledger(cls):
    """Decorate cls with ledger functions."""

    class Ledger(cls):
        """Ledger functions for accountant."""

        def do_get_ledger(self):
            """Return general ledger in Excel."""
            data = self.data
            data = data[0] if data else {}
            attrs = [i for i in data.keys() if not i.startswith("filter_by")]
            start, end = get_interval_dates(data)
            if start:
                comp = self.memory.transaction
                args = ["transaction"]
                args.append(comp.date >= start)
                args.append(comp.date < end)
                items = self.memory.get(*args)
            else:
                items = self.memory.get("transaction")
            r_struct = ReportStruct("ledger", items, self.memory, attrs=attrs)
            self.attachment = ExcelGenerator(r_struct).binary()
            self.attachment_filename = "ledger.xlsx"

        def do_transaction(self):
            """Create transaction in ledger."""
            self.data = self.data[0]
            if "partner" in self.data:
                partner_id = get_partner(self.memory, self.data["partner"])
                self.data["partner"] = partner_id
            self.data["date"] = datetime.date.fromisoformat(self.data["date"])
            self.data["debit_amount"] = float(self.data["debit_amount"])
            if "debit" in self.data:
                self.data["debit"] = int(self.data["debit"])
            if "credit" in self.data:
                self.data["credit"] = int(self.data["credit"])
            booking = Transaction(**self.data)
            self.memory.put(booking)

        def do_recalculate_outstanding(self):
            """Recalculate outstanding entries for a given year and partner."""
            year = int(self.data[0]["year"])
            partner = self.data[0]["partner"]
            for acc in PARTNER_ACCOUNTS:
                debit_stack = get_debitsum_by_partner(self.memory, year, partner, acc)
                credit_stack = get_creditsum_by_partner(self.memory, year, partner, acc)
                for booking in get_partner_bookings(self.memory, year, partner, debit=acc):
                    credit_stack -= booking.debit_amount
                    if round(credit_stack, 2) < 0:
                        booking.debit_stack = round(credit_stack, 2)
                        break
                    booking.debit_stack = 0
                    self.memory.put(booking)
                for booking in get_partner_bookings(
                    self.memory, year, partner, credit=acc
                ):
                    debit_stack -= booking.credit_amount
                    if round(debit_stack, 2) < 0:
                        booking.credit_stack = round(debit_stack, 2)
                        break
                    booking.credit_stack = 0
                    self.memory.put(booking)

    return Ledger



"""Functions for looking up data from memories."""
import dataclasses

from . import memories, token_set_ratio


def get_default_partner(memory):
    """Get default partner."""
    default = memory.get.myself(key="default_partner")
    if not default:
        return None
    return memory.get.partner(id=default.text)


def return_item_val(item):
    """Retun item value based on it's available value."""
    if item.text:
        return item.text
    if item.value:
        return item.value
    if item.dct:
        return item.dct
    if item.lst:
        return item.lst
    if item.blob:
        return item.blob


def get_mydata(memory):
    """Get myself data."""
    myself = memory.get("myself")
    loader = {}
    fields = [i.name for i in dataclasses.fields(memories.Me)]
    for item in myself:
        if item.key in fields:
            loader[item.key] = return_item_val(item)
    return memories.Me(**loader)


def get_partner(memory, text):
    """Return valid partner id from given text and memory access.

    Either finds a partner by text or raises not found First tries text
    to partner index, if that fails performs deep search within Partner
    names
    """
    partner = memory.get(memory.partnerindex.text == text)
    if partner:
        return partner[0].partner_id
    partners = memory.get("partnerindex")
    for i in partners:
        ratio = token_set_ratio(text, i.text)
        if ratio > 95:
            memory.put(memories.PartnerIndex(text, i.partner_id))
            return i.partner_id
    raise RuntimeWarning(f"Partner with name '{text}' could not be found")



"""Functions related to different reports.

Primarily functions on handling Excel reports
"""
import datetime

from .. import handlers, memories


def add_reports(cls):
    """Decorate cls with report functions."""

    class ReportFunctions(cls):
        """Report functions for accountant."""

        def do_get_outstanding(self):
            """Get list of oustanding deals."""
            mem = self.memory
            self.data = self.data[0] if self.data else self.data
            if self.data:
                year = int(self.data.get("year", datetime.date.today().year))
            else:
                year = datetime.date.today().year
            year_start = datetime.date(year=year, month=1, day=1)
            year_end = datetime.date(year=year, month=12, day=31)
            items = []
            filters = (
                (mem.transaction.debit_stack, mem.transaction.debit),
                (mem.transaction.credit_stack, mem.transaction.credit),
            )
            args = [None, None]
            args += [mem.transaction.date >= year_start, mem.transaction.date <= year_end]
            for i in filters:
                # pylint: disable=C0121
                args[0] = i[0] > 0
                for account in [2310, 5310]:
                    args[1] = i[1] == account
                    items += self.memory.get(*args)
            r_struct = memories.ReportStruct("outstanding", items, mem)
            self.attachment = handlers.ExcelGenerator(r_struct).binary()
            self.attachment_filename = "outstanding.xlsx"

    return ReportFunctions



"""Email sending facility."""

import email.message
from smtplib import LMTP


def send(msg: email.message.Message, port: int, host: str = "localhost"):
    """Send email message."""
    with LMTP(host=host, port=port) as lmtp:
        lmtp.send_message(msg, msg["from"], msg["to"])



"""Invoicing module."""
import datetime
from io import BytesIO

from .. import memories, managers, handlers, lookups
from ..templates.invoices import write_invoice, schema
from ..lookups import get_default_partner, get_mydata


def get_invoice_no(partner, memory):
    """Get next invoice number."""
    seq = memory.get.sequences(relation=partner.id)
    if not seq:
        seq = memories.Sequences(relation=partner.id, sequence=1)
        memory.put(seq)
    return partner.invoice_prefix + str(seq.sequence).zfill(3)


def confirm_invoice_no(seq, partner_id, memory):
    """Confirm invoice number."""
    seq += 1
    s = memories.Sequences(relation=partner_id, sequence=seq)
    memory.put(s)


def get_contact_header(name, address, phone=None, email=None):
    """Return valid contact address header."""
    header = [name]
    header += address
    if phone:
        header.append(phone)
    if email:
        header.append(email)
    return header


def get_invoice_bank_info(bank_name, bank_account, bank_code):
    """Return valid invoice bank info."""
    return [bank_name, bank_account, bank_code]


def get_invoice_terms(terms):
    """Return valid invoice terms in days."""
    if not terms:
        return 0
    if terms.isdigit():
        return int(terms)
    raise ValueError(f"Not supported terms format yet {terms}")


def get_invoice_rows(fmt, params):
    """Return valid invoice rows."""
    rows = []
    r_len = len(params)
    today = datetime.date.today()
    cal = today.isocalendar()
    weeks = [i for i in params if i.startswith("week")]
    for i, w in enumerate(weeks):
        params["auto_week_no"] = cal.week - r_len + i
        rows.append(schema.InvoiceRow(fmt.format(**params), params[w]))
    return rows


def get_currency(cur1, cur2):
    """Return valid currency."""
    if cur1:
        return cur1
    if cur2:
        return cur2
    return "EUR"


def add_invoicing(cls):
    """Decorate class with invoicing methods."""

    class Invoicing(cls):
        """Invoicing methods."""

        def do_book_invoice(self):
            """Book invoice."""
            pdf_text = handlers.get_pdf(self.binary)[0]
            invoice = handlers.get_my_invoice(pdf_text)
            path = self.generate_path("out_invoice", "pdf")
            partner_id = lookups.get_partner(self.memory, invoice["partner"])
            booking = memories.Booking("out_invoice", self.memory, **{
                "date": invoice["date"],
                "reference": invoice["invoice_no"],
                "source": path,
                "comment": "sales invoice",
                "debit_amount": float(invoice["amount"]),
                "debit_currency": "EUR",
                "partner": partner_id,
            })
            booking.save()
            handlers.save_file(path, binary=self.binary)
            seq = int(invoice["no"])
            confirm_invoice_no(seq, partner_id, self.memory)

        def do_create_invoice(self):
            """Create new invoice."""
            partner = get_default_partner(self.memory)
            if partner is None:
                self.status = "No default partner set."
                return
            if partner.invoice_prefix is None:
                self.status = "No partner invoice prefix set."
                return
            myself = get_mydata(self.memory)
            addr_mgr = managers.AddressManager(self.memory)
            my_header = get_contact_header(
                myself.name,
                addr_mgr.get(myself.address_id).address_lines,
                myself.phone,
                myself.email)
            payer_header = get_contact_header(
                partner.name,
                addr_mgr.get(partner.address_id).address_lines,)
            my_bank = get_invoice_bank_info(
                myself.bank_name, myself.bank_account, myself.bank_code)
            inv_no = get_invoice_no(partner, self.memory)
            invoice_data = schema.InvoiceData(
                no=inv_no,
                biller=my_header,
                biller_bank=schema.BankInfo(*my_bank),
                payer=payer_header,
                rows=get_invoice_rows(partner.invoice_row_format, self.data[0]),
                terms=get_invoice_terms(partner.terms),
                rate=str(myself.rate),
                currency=get_currency(partner.currency, myself.currency),
            )
            i_bytes = BytesIO()
            write_invoice(i_bytes, invoice_data)
            self.attachment = i_bytes.getvalue()
            self.attachment_filename = f"{inv_no}.pdf"

    return Invoicing



"""Tallybot agent list.

Main idea around tallybot agents is to utilise simple orchestration,
where there is one main agent that must be aware of all other agents and
their capabilities.

Main agent routes tasks as tools to other agents, however other agents
always might move back to main agent if they can't resolve the task.

Main agent can be also considered as a triage agent.
"""

from zoozl.agentgear import BaseAgent, FunctionSchema

from agents import Agent


TallyBotTransfer = FunctionSchema(
    name="transfer_to_Tallybot", description="Transfer to tallybot."
)


class PrivateIncome(BaseAgent):
    """Agent is responsible for handling private income related tasks."""

    instructions = (
        "You book private income."
        "Always answer in sentence or less."
        "Follow the following routine with user.\n"
        "1. Ask for the data.\n"
        " - unless the user has provided it already.\n"
        "2. Book the data with the tool.\n"
        " - unless user wants to do something else.\n"
        "3. Once you have booked the data with your tool, you can transfer to tallybot."
    )
    functions = (
        FunctionSchema(
            name="do_private_income",
            description="book private income entries",
            parameters=[
                {
                    "type": "string",
                    "name": "date",
                    "description": "ISO date of the entry",
                },
                {
                    "type": "number",
                    "name": "amount",
                    "description": "Amount of the entry",
                },
                {
                    "type": "string",
                    "name": "partner",
                    "description": "Name of the partner",
                },
            ],
        ),
        TallyBotTransfer,
    )


class Tallybot(BaseAgent):
    """Responsible for routing tasks to other agents."""

    instructions = (
        "Your name is tallybot. Always be very brief."
        "Your main purpose is to quickly help user to execute any tasks."
        "Tasks can be executed by other agents."
        "If there is no agent that can execute particular task,"
        "apologise that you can't complete it and wait for other tasks."
        "Immediately when you know which task to handle execute it."
    )
    agents = (PrivateIncome,)


tallybot = Agent(
    name="Tallybot",
    instructions=Tallybot.instructions,
)



"""Logical functions performed by accountant."""
import datetime
from html.parser import HTMLParser
import os
import uuid

from .. import memories, handlers, exchange
from ..lookups import get_partner


class Functions():
    """Adds function set to main brain.Perform class."""

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(self, conf, memory, cmd, data, attachment):
        """Init and call executor."""
        self.attachment = None
        self.attachment_filename = None
        self.ecb_xml_memory = None
        self.binary = attachment
        self.data = data
        self.memory = memory
        self.conf = conf
        self.trips = set()
        self.status = "Done"
        getattr(self, cmd)()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_msg, exc_tr):
        pass

    def prepare_trip_report(self, date, tahometer=False):
        """Prepare an excel binary report of trips."""
        title = "tripsummary"
        trips = self.memory.get(
            self.memory.trip.year == date.year,
            self.memory.trip.month == date.month,
        )
        r_struct = memories.ReportStruct(
            title,
            trips,
            self.memory,
            report_args={"tahometer": tahometer}
        )
        self.trips = trips
        self.attachment = handlers.ExcelGenerator(r_struct).binary()
        self.attachment_filename = f"{title}.xlsx"

    def generate_path(self, group, extension, filename=""):
        """Pre-generates file save path."""
        if not filename:
            filename = str(uuid.uuid1()).split('-', maxsplit=1)[0]
        if group not in ["out_invoice", "inc_invoice", "statement"]:
            raise RuntimeWarning(f"Not possible to find a path to save '{group}'")
        path = self.conf[group + "_path"]
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            raise RuntimeWarning(f"Path '{dirname}' does not exist")
        return path + filename + "." + extension

    def exchange_rate(self, date, pair):
        """Give an exchange rate for a date on given pair.

        Otherwise raises Exception.
        """
        date = datetime.date(date.year, date.month, date.day)
        now_date = datetime.date.today()
        msg = "Rates are not available, as requested "
        if date > now_date:
            msg += f"{date} is in future"
            raise RuntimeWarning(msg)
        currency = self.memory.get.currency(date=date, currency=pair)
        if currency:
            return currency
        rate = exchange.get_rate(pair, date)
        currency = memories.Currency(*(date, pair, rate))
        self.memory.put(currency)
        return currency

    def make_expense_booking(self, expense):
        """Do expense transaction."""
        book_type = "expense:"
        if "expense_account" not in expense:
            book_type += "7120"
        else:
            if expense["expense_account"][0] != "7":
                raise RuntimeWarning("Expense account must start with '7'")
            book_type += expense["expense_account"]
        split_expense = False
        if "split" in expense and float(expense["split"]) < 100:
            split_expense = True
            value = round(float(expense["value"]) * float(expense["split"]) / 100, 2)
        else:
            value = float(expense["value"])
        date = datetime.date.fromisoformat(expense["date"])
        ref = expense["reference"]
        partner = get_partner(self.memory, expense.pop("partner"))
        booking = memories.Booking(book_type, self.memory, **{
            "date": date,
            "reference": ref,
            "comment": expense["comment"],
            "partner": partner,
            "source": expense["path"],
            "debit_amount": value,
        })
        booking.save()
        if split_expense:
            booking = memories.Booking("private-expense", self.memory, **{
                "date": date,
                "reference": ref,
                "comment": "private expense",
                "partner": partner,
                "source": expense["path"],
                "debit_amount": round(float(expense["value"]) - value, 2),
            })
            booking.save()


# pylint: disable=W0223
class VIDPageParser(HTMLParser):
    """Parse www.vid.lv page into accounts list."""

    content = False
    div_nest = []
    accounts = []

    def __init__(self, memory):
        """Add memory access."""
        HTMLParser.__init__(self)
        self.memory = memory

    def handle_starttag(self, tag, attrs):
        """As soon as wsite-content div found store it."""
        if tag == "div":
            for i in attrs:
                if i[0] == "id" and i[1] == "wsite-content":
                    self.content = True
        if self.content and tag == "div":
            self.div_nest.append(0)

    def handle_endtag(self, tag):
        """Clean content when wsite-content exits."""
        if tag == "div":
            if self.content and self.div_nest:
                self.div_nest.pop()
            elif self.content:
                self.content = False

    def handle_data(self, data):
        """Parse data and store accounts."""
        if self.content and len(data) > 5 and data[:4].isdigit():
            account = memories.Account(int(data[:4]), data[5:].strip())
            self.memory.put(account)



"""Invoice generator schema."""
from dataclasses import dataclass, field


@dataclass
class InvoiceRow:
    """One invoice row."""

    description: str
    hours: str  # must convert to number
    total: str = field(init=False)


@dataclass
class BankInfo:
    """Bank account information."""

    name: str
    iban: str
    bic: str


@dataclass
class InvoiceData:
    """Invoice creation data container."""

    no: str
    biller: list[str]
    biller_bank: BankInfo
    payer: list[str]
    rows: list[tuple[str, str]]
    terms: int
    rate: str  # must convert to number
    currency: str



"""Managers are objects that allow to interact with memories on a more abstract level.

Managers promise to raise AssertionError with user friendly text upon any error.
"""
from dataclasses import replace

from . import memories
from .lookups import get_partner


class AbstractManager:
    """Abstract manager class."""

    def __init__(self, memory):
        """Initialize with memory access."""
        self.memory = memory


class PartnerManager(AbstractManager):
    """Interact with Partner memory."""

    def __init__(self, memory):
        """Initialize with memory access."""
        super().__init__(memory)
        self.address = AddressManager(memory)

    def update(self, payload):
        """Update partner data."""
        name = payload.pop("name")
        partner_id = get_partner(self.memory, name)
        partner = self.memory.get.partner(id=partner_id)
        if not partner:
            raise AssertionError(f"Partner '{name}' not found")
        address = self.address.update(payload, partner.address_id)
        if address:
            payload["address_id"] = address.id
        if "new_name" in payload:
            self.memory.put(memories.PartnerIndex(name, partner_id))
            payload["name"] = payload.pop("new_name")
        partner = replace(partner, **payload)
        self.memory.put(partner)
        return partner


def addressor(payload, address: memories.Address):
    """Update address data."""
    for i in range(1, 5):
        if f"address{i}" in payload:
            if len(address.address_lines) < i:
                address.address_lines.append(payload.pop(f"address{i}"))
            else:
                address.address_lines[i - 1] = payload.pop(f"address{i}")


class AddressManager(AbstractManager):
    """Interact with Address memory."""

    def extract_valid(self, payload):
        """Extract valid address fields."""
        extract = {}
        for i in range(1, 5):
            if f"address{i}" in payload:
                extract[f"address{i}"] = payload.pop(f"address{i}")
        return extract

    def update(self, payload, address_id):
        """Update address data."""
        changes = self.extract_valid(payload)
        if not changes:
            return None
        address = None
        if address_id:
            address = self.memory.get.address(id=address_id)
        if not address:
            address = memories.Address()
        addressor(changes, address)
        address = replace(address, **changes)
        self.memory.put(address)
        return address

    def get(self, address_id):
        """Get address data."""
        address = self.memory.get.address(id=address_id)
        if address:
            return address
        return memories.Address()



"""Slack related functions.

`send_slack_event` sends a slack event to the server.
`send_challenge` sends a slack challenge to the server.
"""

import hmac
import json
import time
import urllib.request


def get_slack_signature(secret: bytes, body: bytes, timestamp: bytes) -> str:
    """Return slack signature.

    :param secret: Slack signing secret.
    :param body: Request body.
    :param timestamp: Request timestamp in seconds from epoch.
    """
    stamp = b"v0:" + timestamp + b":" + body
    hasher = hmac.new(secret, stamp, digestmod="sha256")
    return "v0=" + hasher.hexdigest()


def send_event(port: int, body: dict, secret: bytes):
    """Send slack event.

    :param port: Port to send slack event to.
    :param body: Slack event body.
    """
    payload = json.dumps(body).encode("utf-8")
    headers = {"Content-Type": "application/json"}
    timestamp = str(int(time.time()))
    headers["X-Slack-Request-Timestamp"] = timestamp
    headers["X-Slack-Signature"] = get_slack_signature(
        secret, payload, timestamp.encode("ascii")
    )
    request = urllib.request.Request(
        f"http://localhost:{port}",
        headers=headers,
        data=payload,
        method="POST",
    )
    try:
        with urllib.request.urlopen(request, timeout=2) as response:
            length = int(response.headers.get("Content-Length", 0))
            return response.status, response.headers, response.read(length)
    except TimeoutError:
        raise AssertionError(f"Request to {request.full_url} timed out")


def send_challenge(challenge: str, port: int, secret: bytes):
    """Send slack challenge."""
    body = {"type": "url_verification", "challenge": challenge}
    return send_event(port, body, secret)



"""Test tallybot command line call."""

import asyncio
import subprocess

import websockets

from tests import base
from tests.simulators import network, slack


class CommandLine(base.TestCase):
    """Test tallybot commandline."""

    def test(self):
        """Test main module."""
        with subprocess.Popen(
            ["env/bin/python", "tallybot", self.config_file],
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            encoding="utf-8",
        ) as subproc:
            try:
                self.connect()
            except Exception:
                subproc.terminate()
                print(subproc.stdout.read())
                raise
            else:
                subproc.terminate()

    def connect(self):
        """Connect to tallybot slack port."""
        port = self.config["slack"]["port"]
        network.wait_for_connection(port)
        challenge = "1234567890"
        status, headers, body = slack.send_challenge(
            challenge, port, self.config["slack"]["signing_secret"].encode("ascii")
        )
        self.assertEqual(status, 200)
        self.assertEqual(body, challenge.encode("ascii"))

    async def connect_websocket(self, port):
        async with websockets.connect(f"ws://localhost:{port}") as websocket:
            await websocket.send('{"text": "Hello, tallybot!"}')
            recv = await asyncio.wait_for(websocket.recv(), timeout=2)
            self.assertEqual(
                recv,
                '{"author": "tallyBot", "text": "What would you like me to do?"}',
            )



"""Module provides external exchange rates."""

import datetime
import urllib.request
import xml.etree.ElementTree as ET


XML_MEMORY: tuple = tuple()  # Holds xml in memory


# pylint: disable=global-statement,too-many-locals
def get_rate(pair, date, base="EUR"):
    """Give an exchange rate for a date on given pair.

    Otherwise raises Exception.
    """
    global XML_MEMORY
    if base != "EUR":
        raise RuntimeError("Only EUR as base currency supported for now")
    day = datetime.timedelta(days=1)
    date = datetime.date(date.year, date.month, date.day) - day
    if not XML_MEMORY or XML_MEMORY[0] < date:
        url = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-sdmx.xml"
        with urllib.request.urlopen(url) as response:
            ext_xml = response.read()
        XML_MEMORY = (datetime.date.today(), ET.fromstring(ext_xml))
    # pylint: disable=too-many-nested-blocks
    rates = {}
    for child in XML_MEMORY[1]:
        tag = child.tag.split("}")
        if len(tag) > 1 and tag[1] == "DataSet":
            for item in child:
                tag = item.tag.split("}")
                if len(tag) > 1 and tag[1] == "Series":
                    curr = item.attrib["CURRENCY"]
                    if curr == pair:
                        for subitems in item:
                            period = subitems.attrib["TIME_PERIOD"]
                            rate = float(subitems.attrib["OBS_VALUE"])
                            rates[period] = rate
    dtfmt = "%Y-%m-%d"
    to_hit = date.strftime(dtfmt)
    if to_hit in rates:
        return rates[to_hit]
    while to_hit[:4] >= "2022":
        # pylint: disable=line-too-long
        to_hit = (
            datetime.datetime.strptime(to_hit, dtfmt)
            - datetime.timedelta(days=1)
        ).strftime(dtfmt)
        if to_hit in rates:
            return rates[to_hit]
    raise RuntimeError(f"Could not find rate for '{pair}' on '{date}'")



# pylint: disable=C0115,C0116
"""Email interface tests for accountant."""
import csv
import datetime
import io
from unittest.mock import patch, MagicMock
import unittest

import membank
from openpyxl import load_workbook

from tallybot import brain, memories
from tests import base


# SMTP mock
smtp_call = MagicMock()
enter_smtp = MagicMock()
enter_smtp.__enter__.return_value = smtp_call
smtp_mock = MagicMock(return_value=enter_smtp)
# Brain Perform mock
brain_call = ["", b"", ""]
brain_mock = MagicMock(return_value=brain_call)
# Brain Perform mock with RuntimeError
brain_mock_exception = MagicMock(side_effect=RuntimeError("oh"))


def get_table(rows):
    """Get from openpyxl rows objec a string table."""
    msg = "\n"
    for row in rows:
        msg += " ".join([str(i.value) for i in row]) + "\n"
    return msg


class PrivateIncome(base.AbstractEmailInterfaceTest):
    """Testcases on doing private income."""

    def setUp(self):
        """Prepare private income email."""
        super().setUp()
        self.email = base.load_email()
        self.email["subject"] = "do private income"
        self.val = 83.18
        partner = "Upwork"
        date = datetime.date(year=2022, month=3, day=24)
        content = f"""date: {date.strftime("%Y-%m-%d")}
            partner: {partner}
            value: {self.val}"""
        self.email.set_content(content)

    def test(self):
        """Make private income entry."""
        self.do_job_call()
        memory = membank.LoadMemory(self.conf["db_path"])
        booking = memory.get(
            memory.transaction.debit == 2310,
            memory.transaction.credit == 8900,
            memory.transaction.debit_amount == self.val,
        )
        self.assertEqual(1, len(booking), list(booking))


class UpdateTransaction(base.AbstractEmailInterfaceTest):
    """Testcase on editing transactions from ledger excel."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.load_partners()
        self.load_seb_statement()

    def test_edit(self):
        """Edit transaction with excel."""
        excel = self.get_excel_ledger()
        sheets = excel.active
        header = True
        for i in sheets:
            if header:
                header = False
                continue
            i[0].value = "2022-03-05"
        tmp = io.BytesIO()
        excel.save(tmp)
        tmp.seek(0)
        excel_bytes = tmp.read()
        tmp.close()
        email = base.load_email(
            fname="ledg", fn_obj=excel_bytes, subtype="xlsx"
        )
        email["subject"] = "do update ledger"
        self.do_job_call(email)
        excel = self.get_excel_ledger()
        new_sheets = excel.active
        for row, new_row in zip(sheets, new_sheets):
            for cell, new_cell in zip(row, new_row):
                self.assertEqual(cell.value, new_cell.value)

    def get_excel_ledger(self):
        email = base.load_email()
        email["subject"] = "do get ledger"
        self.do_job_call(email)
        tmp = io.BytesIO()
        tmp.write(self.response[1])
        tmp.seek(0)
        excel = load_workbook(tmp)
        tmp.close()
        return excel


class DeleteTransaction(base.AbstractEmailInterfaceTest):
    """Testcase on deleting transactions."""

    def test_delete(self):
        """Calling delete should delete a transaction."""
        mem = self.memory
        self.load_transactions()
        trans = mem.get("transaction")
        self.assertTrue(len(trans) > 0)
        for i in trans:
            email = base.load_email(content=f"id: {i.id}")
            email["subject"] = "do delete transaction"
            self.do_job_call(email)
        trans = mem.get("transaction")
        self.assertEqual(len(trans), 0)


class Assets(base.AbstractEmailInterfaceTest):
    """Testcase on asset depreciation."""

    def test_asset_calculation(self):
        mem = self.memory
        val = 1000
        content = f"""
            date: 2022-03-22
            name: Laptop
            reference: 5698934
            value: {val}
            utility: 23
        """
        email = base.load_email(
            content=content, fname="asset", subtype="pdf", fn_obj=1
        )
        email["subject"] = "do add initial asset"
        self.do_job_call(email, assert_status=False)
        books = mem.get(mem.transaction.debit == 7420)
        self.assertEqual(len(books), 23)
        for i in books:
            self.assertEqual(round(val / 23, 2), i.debit_amount)


class SocialReport(base.AbstractEmailInterfaceTest):
    """Testcase on social tax report."""

    @patch("accountant.brain.main.handlers.save_file", new=MagicMock())
    def test_report(self):
        self.load_transactions()
        self.load_upwork_invoices()
        email = base.load_email(content="quarter: 2022-Q1")
        email["subject"] = "do get social report"
        self.do_job_call(email, assert_status=False)
        self.assertIn("Social report", self.response[0])


class CarWash(base.AbstractEmailInterfaceTest):
    """Testcases on carwash receipts."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.memory = membank.LoadMemory(self.conf["db_path"])
        self.load_partners()

    @patch("accountant.brain.main.handlers.save_file")
    def test_wash(self, save_file):
        self.do_add_carwash()
        expense = self.memory.get(self.memory.transaction.debit == 7760)
        self.assertEqual(len(expense), 1)
        save_file.assert_called()


class ManageTrips(base.AbstractEmailInterfaceTest):
    """Testcases on car trips."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.email = base.load_email(fname="receipt", subtype="pdf")
        self.load_partners()
        self.load_seb_statement()
        self.memory = membank.LoadMemory(self.conf["db_path"])
        self.excel_rows = {}
        self.args = {
            "date": "2022-03-05",
            "trip": "Rīga - Madona",
            "description": "Maintenance of server room",
            "distance": 230,
        }

    def test_trip_errors(self):
        date = "date: 2022-03-24\n"
        dist = "distance: 200\n"
        desc = "description:\n"
        trip = "trip:\n"
        ref = "receipt_ref: some\n"
        val = "receipt_value:\n"
        lit = "receipt_litres\n"
        cases = []
        cases.append([date, dist, desc])
        cases.append([date, dist, trip])
        cases.append([date + ref, val])
        cases.append([date + ref, lit])
        for i in cases:
            self.error_checks(i)
        email = base.load_email()
        email["subject"] = "do trip summary"
        self.do_job_call(email, assert_status=False)
        self.assertIn("There is no date found", self.response[0])

    def error_checks(self, args, flag="missing"):
        content = ""
        for i in args:
            content += i
            email = base.load_email(
                content=content, fname="receipt", subtype="pdf"
            )
            email["subject"] = "do add trip"
            self.do_job_call(email, assert_status=False)
            if flag == "missing":
                self.assertIn(
                    "Missing required field", self.response[0], content
                )

    @patch("accountant.brain.main.handlers.save_file")
    def test_trip_booking(self, save_file):
        self.args["receipt_ref"] = "576"
        self.args["receipt_litres"] = 67
        self.args["receipt_value"] = 68
        self.make_trip(**self.args)
        save_file.assert_called_once()
        self.assertIn(
            "tests/tmp/ienakosie_rekini/", save_file.call_args.args[0]
        )
        self.assertEqual(b"0", save_file.call_args.args[1])
        email = base.load_email()
        email["subject"] = "do get carinfo"
        self.do_job_call(email, assert_status=False)
        self.assertIn(str(memories.DEFAULT_CAR_ODO_KM + 5), self.response[0])

    def test_add_trip(self):
        args = self.args
        self.make_trip(**args)
        self.check_trip_summary()
        tests = (
            (args["date"], "A11"),
            (None, "B11"),
            (None, "C11"),
            (args["trip"], "D11"),
            (args["description"], "E11"),
            (args["distance"], "F11"),
        )
        self.assert_values(tests)

    @patch("accountant.brain.main.handlers.save_file")
    def test_add_trips(self, save_file):
        total_distance = self.make_trips()[0]
        self.check_trip_summary()
        tests = (
            ("2022-03-01", "A11"),
            ("2022-03-04", "A14"),
        )
        self.assert_values(tests)
        distance = sum(self.excel_rows["F" + str(11 + i)] for i in range(5))
        self.assertEqual(total_distance, distance)
        car = self.memory.get("car")[0]
        self.assertEqual(car.total_km, total_distance)
        save_file.assert_called()

    def assert_values(self, values):
        for value, cell in values:
            msg = f"{value} not found in {cell}"
            self.assertEqual(value, self.excel_rows[cell], msg)

    def test_get_trip(self):
        self.check_trip_summary()
        self.check_trip_summary()

    @patch("accountant.brain.main.handlers.save_file", new=MagicMock())
    def test_fuel_expense(self):
        total_distance, litres = self.make_trips(9)
        split = total_distance * 8 / 100 / litres
        content = "date: 2022-03\n"
        content += "tahometer: 1800000\n"
        content += "force:"
        email = base.load_email(fn_obj=1, content=content)
        email["subject"] = "do add fuel expense"
        self.do_job_call(email)
        self.assertTrue(self.response[1])
        fuel_booking = self.memory.get(self.memory.transaction.debit == 7760)
        self.assertEqual(len(fuel_booking), 4)
        for booking in fuel_booking:
            self.assertEqual(booking.credit, 5310)
            self.assertEqual(booking.debit_amount, round(100 * split, 2))
            self.assertEqual(booking.debit_amount, booking.credit_amount)

    def test_fuel_expense_force(self):
        content = ""
        for i, val in enumerate(["date: 2022-01\ntahometer: 0\n", "force:"]):
            content += val
            email = base.load_email(fn_obj=1, content=content)
            email["subject"] = "do add fuel expense"
            self.do_job_call(email, assert_status=False)
            if i == 0:
                self.assertIn("first 5 days of next month", self.response[0])
            else:
                self.assertIn("There are no fuel receipts", self.response[0])

    def make_trip(self, **kargs):
        lit = kargs.get("receipt_litres", "")
        ref = kargs.get("receipt_ref", "")
        val = kargs.get("receipt_value", "")
        if any((lit, ref, val)):
            content = f"""
                date: {kargs["date"]}
                receipt_ref: {ref}
                receipt_litres: {lit}
                receipt_value: {val}
            """
        else:
            content = f"""
                date: {kargs["date"]}
                trip: {kargs["trip"]}
                description: {kargs["description"]}
                distance: {kargs["distance"]}
            """
        email = base.load_email(
            fname="receipt", subtype="pdf", fn_obj=1, content=content
        )
        email["subject"] = "do add trip"
        self.do_job_call(email)

    def make_trips(self, count=6):
        total_distance = total_litres = 0
        keys = ["trip", "description", "distance"]
        for i in range(1, count):
            trip = dict(zip(keys, ["something"] * len(keys)))
            trip["date"] = f"2022-03-0{i}"
            if i % 2 == 0:
                trip["distance"] = "5"
                trip["receipt_litres"] = f"{i * 4}"
                trip["receipt_value"] = 100
                trip["receipt_ref"] = "something"
            else:
                trip["distance"] = f"10{i}"
            self.make_trip(**trip)
            total_distance += int(trip["distance"])
            if "receipt_litres" in trip:
                total_litres += float(trip["receipt_litres"])
        return total_distance, total_litres

    def check_trip_summary(self, date="2022-03"):
        email = base.load_email(content=f"date: {date}")
        email["subject"] = "do trip summary"
        self.do_job_call(email)
        rows = self.get_excel_rows()
        coords = {
            i[j].coordinate: i[j].value for i in rows for j in range(len(i))
        }
        self.assertIn("Business Trip Report", coords["D4"])
        self.assertIn("Date", coords["A3"])
        self.assertIn("Fuel: diesel", coords["A5"])
        self.assertIn("Consumption: 8l/100km", coords["A6"])
        self.assertIn("Total", coords["E9"])
        self.assertIn("SUM", coords["F9"])
        self.assertIn("Date", coords["A10"])
        self.excel_rows = coords


class ManageTransactions(base.AbstractEmailInterfaceTest):
    """Testcase to delete transaction."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.email = base.load_email()
        self.load_partners()
        self.load_seb_statement()
        self.memory = membank.LoadMemory(self.conf["db_path"])

    def test_delete(self):
        self.email["subject"] = "do remove transaction"
        bookings_delete = self.memory.get("transaction")
        self.assertTrue(len(bookings_delete) > 0)
        for i in bookings_delete:
            self.email.set_content(f"id: {i.id}")
            self.do_job_call(self.email)
        bookings_delete = self.memory.get("transaction")
        self.assertTrue(len(bookings_delete) == 0, bookings_delete)


class OutstandingDeals(base.AbstractEmailInterfaceTest):
    """Testcase to get outstanding deals."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.email = base.load_email()
        self.email["subject"] = "do get outstanding"
        self.email.set_content("year: 2022")
        self.load_partners()

    def test(self):
        self.load_transactions()
        self.do_job_call(self.email)
        rows = self.get_excel_rows()
        self.assertEqual(36, len(list(rows)))
        header = True
        for i in rows:
            if header:
                header = False
                self.assertEqual(i[0].value, "Date")
                self.assertEqual(i[2].value, "Partner")
                continue
            self.assertEqual(len(i), 12)

    def test_upwork_matching(self):
        self.load_upwork_invoices()
        self.load_upwork_statement()
        self.do_job_call(self.email)
        rows = self.get_excel_rows()
        self.assertEqual(len(list(rows)), 1, get_table(rows))

    def test_seb_matching(self):
        self.load_seb_statement()
        self.do_job_call(self.email)
        rows = self.get_excel_rows()
        self.assertEqual(len(list(rows)), 24, get_table(rows))

    def test_split_expense(self):
        """Both private and business part should be matched."""
        self.do_add_expense({"split": 50, "value": 50})
        self.do_job_call(self.email)
        rows = self.get_excel_rows()
        self.assertEqual(len(list(rows)), 3, get_table(rows))


class Expenses(base.AbstractEmailInterfaceTest):
    """Testcase to get expenses."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.email = base.load_email()
        self.email["subject"] = "do add expense"
        self.load_partners()
        self.load_seb_statement()

    def test_wrong_expense_account(self):
        content = {"expense_account": 6570}
        self.do_add_expense(content)
        self.assertIn("must start with '7'", self.response[0])

    def test_private_expense(self):
        del self.email["subject"]
        self.email["subject"] = "do private expense"
        val = 83.18
        partner = "Klements un Peteris"
        date = datetime.date(year=2022, month=3, day=24)
        content = f"""date: {date.strftime("%Y-%m-%d")}
            partner: {partner}
            Value: {val}"""
        self.email.set_content(content)
        self.do_job_call(self.email)
        memory = membank.LoadMemory(self.conf["db_path"])
        expense = memory.get(
            memory.transaction.debit == 8900,
            memory.transaction.debit_amount == val,
        )
        self.assertEqual(1, len(expense), list(expense))
        expense = expense[0]
        self.assertEqual(expense.credit, 5310)
        self.assertEqual(expense.date, date)
        self.assertEqual(expense.debit_amount, val)
        self.assertEqual(expense.debit_amount, expense.credit_amount)

    def test(self):
        ref = "T394847585"
        val = 83.18
        comment = "Expense for buying retro"
        date = datetime.date(year=2022, month=3, day=23)
        partner = "Circle K"
        content = {
            "reference": ref,
            "comment": comment,
            "value": val,
            "partner": partner,
            "date": date.isoformat(),
        }
        self.do_add_expense(content)
        memory = membank.LoadMemory(self.conf["db_path"])
        expense = memory.get(memory.transaction.debit == 7120)
        self.assertEqual(1, len(expense))
        expense = expense[0]
        self.assertEqual(val, expense.debit_amount)
        self.assertEqual(expense.debit_amount, expense.credit_amount)
        self.assertEqual(0, expense.credit_stack)
        self.assertEqual(ref, expense.reference)
        self.assertEqual(date, expense.date)
        self.assertEqual(comment, expense.comment)
        self.assertEqual(5310, expense.credit)

    def test_split(self):
        val = 83.18
        split = 33
        ref = "splitted_expense_test"
        comment = "Expense with not 100% business"
        content = {
            "reference": ref,
            "comment": comment,
            "value": val,
            "partner": "Circle K",
            "split": split,
        }
        self.do_add_expense(content)
        memory = membank.LoadMemory(self.conf["db_path"])
        expense = memory.get(memory.transaction.reference == ref)
        self.assertEqual(2, len(expense))
        self.assertIn(8900, [i.debit for i in expense])
        self.assertIn(7120, [i.debit for i in expense])
        self.assertEqual(val, round(sum(i.debit_amount for i in expense), 2))
        for i in expense:
            self.assertEqual(0, i.credit_stack)
            self.assertEqual(i.debit_amount, i.credit_amount)
            if i.debit == 8900:
                self.assertEqual(
                    round(val * (100 - split) / 100, 2), i.debit_amount
                )
            else:
                self.assertEqual(
                    round(val * (split) / 100, 2), i.debit_amount, i
                )
            self.assertEqual(5310, i.credit, i)

    def test_split_100(self):
        val = 89.0
        split = 100
        ref = "splitted_expense_test"
        content = {
            "reference": ref,
            "value": val,
            "partner": "Circle K",
            "split": split,
        }
        self.do_add_expense(content)
        memory = membank.LoadMemory(self.conf["db_path"])
        expense = memory.get(memory.transaction.reference == ref)
        self.assertEqual(1, len(expense))


class Partners(base.AbstractEmailInterfaceTest):
    """Testcase to get partners."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.email = base.load_email()

    def create_partner(self, name="Funny"):
        self.email["subject"] = "do create partner"
        self.email.set_content(f"name: {name}")
        self.do_job_call(self.email, assert_status=False)

    def test_load_partners(self):
        self.load_partners()
        memory = membank.LoadMemory(self.conf["db_path"])
        partners = memory.get("partner")
        self.assertEqual(len(partners), 17)
        p_ids = {i.id for i in partners}
        self.assertEqual(len(p_ids), 17)

    def test_creation(self):
        self.create_partner()
        memory = membank.LoadMemory(self.conf["db_path"])
        partners = memory.get("partner")
        self.assertEqual(len(partners), 1)

    def test_update(self):
        self.create_partner("Funny K")
        self.email = base.load_email()
        self.email["subject"] = "do update partner"
        self.email.set_content("name: Funny K")
        response = self.do_job_call()
        memory = membank.LoadMemory(self.conf["db_path"])
        partner = memory.get("partner")[0]
        self.assertEqual("Funny K", partner.name)
        content = response[0].split("\n")
        content[0] = "name: FunnyDouble"
        content = "\n".join(content)
        self.email.set_content(content)
        del self.email["subject"]
        self.email["subject"] = "RE: do update partner"
        response = self.do_job_call()
        self.assertIn("Done", response[0])
        updated_partner = memory.get(memory.partner.id == partner.id)[0]
        self.assertEqual(updated_partner.name, "FunnyDouble")

    def test_get_partners(self):
        self.create_partner()
        self.email = base.load_email()
        self.email["subject"] = "do list partners"
        response = self.do_job_call()
        self.assertEqual(response[0], "Done")
        self.assertTrue(response[1])


class GetLedger(base.AbstractEmailInterfaceTest):
    """General ledger report tests."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.load_transactions()

    def make_ledger_email(self, content=""):
        email = base.load_email(content=content)
        email["subject"] = "do get ledger"
        self.email = email

    def test_ledger(self):
        """Receive attachment as Excel report."""
        self.make_ledger_email()
        self.do_job_call(self.email)
        rows = self.get_excel_rows()
        header = True
        counter = 0
        for i in rows:
            counter += 1
            if header:
                header = False
                self.assertEqual(i[0].value, "Date")
                self.assertEqual(i[2].value, "Partner")
                continue
            self.assertEqual(len(i), 12, [i.value for i in i])
            date = i[0].value
            self.assertTrue(date[:4].isdigit(), date)
            self.assertTrue(i[1].value, "ledger misses reference")
            self.assertEqual(len(i[11].value), 40, "ledger misses id")
        self.assertEqual(counter, 43)

    def test_get_specific_fields(self):
        content = """
            date:
            debit_stack:
            id:
        """
        self.make_ledger_email(content)
        self.do_job_call(self.email)
        rows = self.get_excel_rows()
        for row in rows:
            self.assertEqual(len(row), 3, [i.value for i in row])

    def test_get_specific_month(self):
        content = """
            filter_by_month: 2022-03
            date:
            partner:
            id:
        """
        all_books = self.memory.get("transaction")
        books = self.memory.get(
            self.memory.transaction.date
            >= datetime.date(year=2022, month=3, day=1),
            self.memory.transaction.date
            < datetime.date(year=2022, month=4, day=1),
        )
        self.assertNotEqual(all_books, books)
        self.make_ledger_email(content)
        self.do_job_call(self.email)
        rows = self.get_excel_rows()
        counter = -1  # discount header row
        for row in rows:
            self.assertEqual(len(row), 3, [i.value for i in row])
            counter += 1
        self.assertEqual(len(books), counter)


class ExchangeRateFetch(base.AbstractEmailInterfaceTest):
    """Verifies that exchange rate is always available."""

    def test(self):
        """Check rate available for yesterday."""
        now = datetime.date.today()
        day = datetime.timedelta(days=1)
        with brain.Perform(
            self.conf, self.memory, "do_get_help", {}, b""
        ) as job:
            job.exchange_rate(now - day, "USD")
        with brain.Perform(
            self.conf, self.memory, "do_get_help", {}, b""
        ) as job:
            rate = job.exchange_rate(now, "USD")
        with brain.Perform(
            self.conf, self.memory, "do_get_help", {}, b""
        ) as job:
            rate90 = job.exchange_rate(now - 90 * day, "USD")
        self.assertTrue(rate)
        with brain.Perform(
            self.conf, self.memory, "do_get_help", {}, b""
        ) as job:
            new_rate = job.exchange_rate(now, "USD")
        self.assertEqual(rate, new_rate)
        self.assertNotEqual(rate90, new_rate)

    def test_wrong_input(self):
        """Ask rates in future or too away in past."""
        now = datetime.date.today()
        day = datetime.timedelta(days=1)
        scenarios = ((now + day, "is in future"),)
        for i in scenarios:
            with self.assertRaises(RuntimeWarning) as exc:
                with brain.Perform(
                    self.conf, self.memory, "do_get_help", {}, b""
                ) as job:
                    job.exchange_rate(i[0], "USD")
                self.assertIn(i[1], str(exc))


class SebStatement(base.AbstractEmailInterfaceTest):
    """Check that SEB statement is uploaded."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.load_partners()

    def test(self):
        """Upload seb statement."""
        self.load_seb_statement()
        memory = membank.LoadMemory(self.conf["db_path"])
        transactions = memory.get("transaction")
        transactions = {i.reference: i for i in transactions}
        tests = [
            ("RO1008215702L01", "Kafijas Draugs", 45.97, 5310),
            ("RO1002210337L01", "apkalpošanas maksa", 2.0, 5310),
        ]
        for test in tests:
            self.assertIn(test[0], transactions)
            expect = transactions[test[0]]
            self.assertIn(test[1], expect.comment, expect)
            self.assertEqual(test[2], expect.debit_amount, expect)
            self.assertEqual(test[2], expect.credit_amount, expect)
            self.assertEqual(test[3], expect.debit, expect)
            self.assertEqual("EUR", expect.debit_currency, expect)
            self.assertEqual("EUR", expect.credit_currency, expect)
            self.assertTrue(expect.partner, expect)


class UpworkInvoices(base.AbstractEmailInterfaceTest):
    """Check that Upwork invoices are uploaded."""

    def setUp(self):
        base.AbstractEmailInterfaceTest.setUp(self)
        self.load_partners()
        self.email = base.load_email(
            fname="test_upwork_invoices", subtype="zip"
        )

    def test(self):
        """Upload zipped PDF invoices."""
        self.email["subject"] = "do upwork invoices"
        memory = self.memory
        self.assertEqual(0, len(memory.get("transaction")))
        response = self.do_job_call()
        self.assertEqual(response[0], "Done")
        items = memory.get("transaction")
        base.remove_temps()
        response = self.do_job_call()
        self.assertEqual(response[0], "Done")
        self.assertEqual(len(items), len(memory.get("transaction")))
        items = {
            i.reference: i for i in items if self.assertTrue(i.partner) is None
        }
        expected = ["T463972580", "T462044954", "T463972580", "T460107076"]
        for i in expected:
            self.assertIn(i, items, items.keys())
        self.assertEqual(items["T462044954"].comment, "Service Fee")
        self.assertEqual(
            items["T463972580"].comment,
            "(27898219) Juris Kaminskis - 24:00 hrs @ $30.00/hr",
        )
        book_test = items["T463972580"]
        partner = memory.get.partner(id=book_test.partner)
        self.assertTrue(partner)
        self.assertEqual(partner.name, "UNIT9")
        self.assertEqual(book_test.debit, 2310)
        self.assertEqual(book_test.credit, 6110)
        self.assertEqual("T463972580", book_test.reference)
        book_test = items["T460107076"]
        partner = memory.get.partner(id=book_test.partner)
        self.assertTrue(partner, book_test.partner)
        self.assertEqual(partner.name, "Upwork")
        self.assertEqual(book_test.debit, 7170)
        self.assertEqual(book_test.credit, 5310)
        self.assertEqual("T460107076", book_test.reference)
        nodes = memory.get(memory.node.transaction == book_test.id)
        self.assertEqual(len(nodes), 2)
        for node in nodes:
            if node.account == 5310:
                self.assertEqual(node.closing, -9.07)
            else:
                self.assertEqual(node.closing, 92.57)


class Accounts(base.AbstractEmailInterfaceTest):
    """Check that accounts are uploaded."""

    # TODO: reimplement accounts update
    @unittest.skip("Reimplement accounts update")
    def test(self):
        """Send request to upload accounts."""
        self.email["subject"] = "do accounts update"
        response = self.do_job_call()
        self.assertEqual(response[0], "Done")
        memory = membank.LoadMemory(self.conf["db_path"])
        account = memory.get.account()
        self.assertEqual(account.id, 1110)
        self.assertEqual(
            account.name, "Pētniecības un uzņēmuma attīstības izmaksas"
        )


class UpworkStatement(base.AbstractEmailInterfaceTest):
    """Check that Upwork statement is uploaded."""

    @patch("accountant.brain.main.handlers.save_file", new=MagicMock())
    def test(self):
        """Basic upload."""
        self.load_partners()
        self.load_upwork_invoices()
        self.load_transactions()
        memory = self.memory
        items = memory.get("transaction")
        items = {i.reference: i for i in items}
        with open(
            "tests/data/test_upwork_statement", encoding="utf-8"
        ) as fname:
            csv_reader = csv.reader(fname, delimiter=",")
            header = True
            for row in csv_reader:
                if header:
                    header = False
                    continue
                self.assertIn(row[1], items)
                self.assertIn("upwork", items[row[1]].source)
                orig_amount = items[row[1]].debit_amount
                amount = items[row[1]].deal_value
                self.assertEqual(abs(orig_amount), orig_amount)
                self.assertEqual(abs(amount), amount)
                if "Service Fee" in items[row[1]].comment:
                    self.assertEqual(items[row[1]].debit, 5310, items[row[1]])
                    self.assertEqual(items[row[1]].credit, 2670, items[row[1]])
                    self.assertEqual(items[row[1]].credit, 2670, items[row[1]])
                    self.assertTrue(items[row[1]].partner, items[row[1]])
                elif "Invoice for" in items[row[1]].comment:
                    self.assertEqual(items[row[1]].debit, 2670, items[row[1]])
                    self.assertEqual(items[row[1]].credit, 2310, items[row[1]])
                    self.assertTrue(items[row[1]].partner, items[row[1]])
                elif "Withdrawal Fee" in items[row[1]].comment:
                    self.assertEqual(items[row[1]].debit, 7170, items[row[1]])
                    self.assertEqual(items[row[1]].credit, 2670, items[row[1]])
                else:
                    self.assertEqual(items[row[1]].debit, 2680, items[row[1]])
                    self.assertEqual(items[row[1]].credit, 2670, items[row[1]])
        self.assert_withdrawal_booking()

    def assert_withdrawal_booking(self):
        memory = self.memory
        upwork_send = memory.get(memory.transaction.debit == 2680)[0]
        seb_receive = memory.get(memory.transaction.credit == 2680)[0]
        exch_loss = upwork_send.deal_value - seb_receive.debit_amount
        exch_book = memory.get(memory.transaction.credit == 8150)
        self.assertEqual(len(exch_book), 1)
        exch_book = exch_book[0]
        self.assertEqual(exch_book.debit_amount, 19.56, exch_book)
        book = memory.get(
            memory.transaction.debit == 7170,
            memory.transaction.credit == 2670,
            memory.transaction.debit_amount == round(exch_loss, 2),
        )
        self.assertEqual(len(book), 1, book)



"""Functions to handle various email attachment types and data conversions."""

import datetime
import io
import json
import os
import re
import string
import zipfile

import pdftotext  # type: ignore[import-not-found]
import openpyxl  # typpe: ignore


def get_content(msg):
    """Get content from email msg type.

    Return content as as tuple of (data, attachment)
    """
    data = {}
    attachment = b""
    supported = ["csv", "zip", "xlsx", "pdf"]
    for i in msg.walk():
        cur_type = i.get_content_subtype()
        if cur_type in supported or cur_type == "plain":
            content = i.get_content()
            if "plain" == cur_type:
                data = get_json(content)
            elif cur_type in ["csv", "zip", "xlsx", "pdf"]:
                attachment = content
    return data, attachment


# pylint: disable=R0903
class ExcelGenerator:
    """Generates Excel."""

    def __init__(self, *r_structs):
        """From Report structure will generate Excel binary."""
        self.__r_structs = r_structs
        self.__write_excel_sheets()

    def __write_excel_sheets(self):
        """Convert r_structs to Excel rows format."""
        self.__sheets = []
        for r_struct in self.__r_structs:
            rows = []
            rows += r_struct.header
            start = r_struct.start
            for i, j in enumerate(r_struct.attrs):
                rows.append(
                    (f"{string.ascii_uppercase[i]}{str(start)}", j.capitalize())
                )
            for a, item in enumerate(r_struct.items):
                for i, j in enumerate(r_struct.attrs):
                    val = getattr(item, j)
                    if j in r_struct.callbacks:
                        try:
                            val = r_struct.callbacks[j](val)
                        except AttributeError:
                            raise RuntimeWarning(
                                f"Transaction {item} misses partner"
                            ) from None
                    rows.append(
                        (
                            f"{string.ascii_uppercase[i]}{str(start + a + 1)}",
                            val,
                        )
                    )
            self.__sheets.append((rows, r_struct.title))

    def binary(self):
        """Prepare binary stream object.

        Requires a rows (cell data to be written in tuple (cell, data))
        That is prepared into Excel file and returned as binary file-
        object.
        """
        excel = openpyxl.Workbook()
        first = True
        for sheet in self.__sheets:
            if first:
                ws_sheet = excel.active
                ws_sheet.title = sheet[1]
                first = False
            else:
                ws_sheet = excel.create_sheet(title=sheet[1])
            for cell in sheet[0]:
                ws_sheet[cell[0]].value = cell[1]
        tmp = io.BytesIO()
        excel.save(tmp)
        tmp.seek(0)
        excel_bytes = tmp.read()
        tmp.close()
        return excel_bytes


def get_excel(content):
    """From bytes content returns Openpyxl object."""
    bin_read = io.BytesIO()
    bin_read.write(content)
    bin_read.seek(0)
    excel = openpyxl.load_workbook(bin_read)
    bin_read.close()
    return excel


def get_json(text):
    """From plain text 'key:value format' builds json string."""
    content = text.split("\n")
    json_string = "["
    data = "{"
    previous = False
    header = True
    for i, j in enumerate(content):
        col = j.find(":")
        if col == -1:
            continue
        val = j[col + 1 :].strip()
        key = j[0:col].strip().lower()
        if key == previous:
            json_string += data[:-1] + "},"
            data = "{"
        data += f'"{key}": "{val}",'
        if i == len(content) - 1:
            json_string += data[:-1] + "}"
            data = ""
        if header:
            previous = key
            header = False
    if data and data[-1] == ",":
        json_string += data[:-1] + "}"
    json_string += "]"
    try:
        json_string = json.loads(json_string)
    except json.JSONDecodeError:
        raise RuntimeWarning(f"Corrupt json format {json_string}")
    return json_string


def get_zip(content):
    """Return a list of all files in zip archive."""
    fn = io.BytesIO()
    fn.write(content)
    fn.seek(0)
    with zipfile.ZipFile(fn) as archive:
        result = []
        for i in archive.namelist():
            with archive.open(i) as temp:
                result.append(temp.read())
    fn.close()
    return result


def get_pdf(content):
    """Return pdf object."""
    fn = io.BytesIO(content)
    # pylint: disable=I1101
    pdf = pdftotext.PDF(fn, physical=True)
    fn.close()
    return pdf


def get_my_invoice(text):
    """Return invoice data from text."""
    patterns = {
        "invoice_no": r"INVOICE #\s*(\w+)",
        "date": r"INVOICE DATE\s*(.+)",
        "amount": r"TOTAL AMOUNT\s*EUR\s*(\d*,*\d+\.\d{2})",
        "partner": r"Bill to:\s*(.+)",
    }
    for key, p in patterns.items():
        res = re.search(p, text)
        if res:
            patterns[key] = res.group(1)
        else:
            patterns[key] = ""
    patterns["no"] = re.search(r"\d+", patterns["invoice_no"]).group()
    patterns["amount"] = patterns["amount"].replace(",", "")
    patterns["date"] = datetime.datetime.strptime(
        patterns["date"].replace(" ", ""), "%d%b%Y"
    )
    return patterns


# pylint: disable=R0912,R0915
def get_invoice(text):
    """Return invoice data from text."""
    invoice_date = reference = comment = value = issuer = partner = ""
    for i, j in enumerate(text):
        if j == "F":
            char = text[i : i + 5]
            if char == "From:":
                issuer = text[i + 6 : i + 100].strip().split("\n")[0]
        elif j == "B":
            if text[i : i + 8] == "Bill to:":
                partner = text[i + 9 : i + 50].strip()
        elif j == "I":
            char = text[i : i + 9]
            if char == "INVOICE #":
                word_start = False
                word_end = False
                word = ""
                j = 0
                while not all((word_start, word_end)) and j < 50:
                    j += 1
                    char = text[i + 9 + j].strip()
                    if char:
                        word += char
                        word_start = True
                    if not char and word_start:
                        word_end = True
                reference = word.strip()
        elif j == "D":
            char = text[i - 5 : i + 8]
            if char == "     DATE    ":
                invoice_date = text[i + 4 : i + 42].strip()
            char = text[i : i + 18]
            if char == "DESCRIPTION / MEMO":
                a = 18
                while True:
                    char = text[i + a]
                    if comment == "AMOUNT":
                        comment = ""
                        break
                    if len(comment) > 150:
                        raise RuntimeWarning(
                            "Could not find properly comment field"
                        )
                    comment += char
                    comment = comment.strip()
                    a += 1
                comment = text[i + a : i + a + 150]
                comment = comment.strip()
                comment = comment.split("   ")[0]
        elif j == "T":
            char = text[i - 5 : i + 17]
            if char == " " * 5 + "TOTAL AMOUNT" + " " * 5:
                value = text[i + 12 : i + 39].strip()[1:]
    if not all((invoice_date, reference, comment, value, issuer, partner)):
        err_msg = "Could not fetch all required data from invoice\n"
        err_msg += f"Invoice date: {invoice_date}\n"
        err_msg += f"Reference: {reference}\n"
        err_msg += f"Comment: {comment}\n"
        err_msg += f"Value: {value}\n"
        err_msg += f"Issuer: {issuer}\n"
        err_msg += f"Partner: {partner}\n\n{text}"
        raise RuntimeWarning(err_msg)
    invoice_date = datetime.datetime.strptime(invoice_date, "%b %d, %Y")
    value = value.replace(",", "")
    return (
        invoice_date,
        reference,
        comment,
        value,
        issuer,
        partner,
    )


def save_file(full_path, binary):
    """Save binary file as per full_path."""
    if os.path.exists(full_path):
        raise RuntimeWarning(f"File already exists in {full_path}")
    with open(full_path, mode="wb") as fname:
        fname.write(binary)



"""Common classes and methods for tests."""

import email
import datetime
from dataclasses import dataclass
import logging
import io
import os
import random
import tomllib
from types import MappingProxyType
import unittest
import uuid

import membank
import openpyxl

from tallybot import brain
from zoozl.tests import TestEmailbot


TEMP_STORAGE_PATH = "tests/tmp/"
logging.basicConfig(level=logging.WARNING)

class TestCase(unittest.IsolatedAsyncioTestCase):
    """Base test case."""

    config_file = "tests/config.toml"

    @classmethod
    def setUpClass(cls):
        """Setup testcase with configuration."""
        cls.config = cls.load_configuration(cls.config_file)

    @staticmethod
    def load_configuration(config_file):
        """Load configuration."""
        with open(config_file, "rb") as f:
            c = tomllib.load(f)
            c["extensions"] = ["tallybot.plugin"]
            return MappingProxyType(c)


def unique_name(self=None):
    """Return unique name."""
    name = ""
    if self:
        cls = getattr(self, "__class__", False)
        if cls:
            name = getattr(cls, "__name__", "Unknown")
    if not name:
        name = "Random"
    return name + "_" + str(uuid.uuid1()).split("-", maxsplit=1)[0]


def random_amount():
    """Return random decimal amount."""
    base = random.random()
    return base * 10 ** random.randrange(10)


def correct_dates(fname, output):
    """Correct dates in testdata files.

    testdata files were generated months ago and dates for
    them must be corrected to sooner ones

    fname - testdata file name
    output - file content read by readlines()
    """
    today = datetime.date.today()
    cur_month = datetime.date(year=today.year, month=today.month - 1, day=1)
    cur_month = cur_month.strftime("%b")
    prev_month = datetime.date(year=today.year, month=today.month - 2, day=1)
    prev_month = prev_month.strftime("%b")
    if "test_upwork_statement" == fname:
        for i, line in enumerate(output[1:]):
            if line[1:4] == "Apr":
                line = line[0] + prev_month + line[4:]
            else:
                line = line[0] + cur_month + line[4:]
            output[i + 1] = line
    else:
        print(fname)
    return "".join(output)


def load_email(fname=False, subtype=False, fn_obj=0, content=""):
    """Load test email.

    If fname and subtype present loads a fname as attachment to email
    message.
    """
    path = "tests/data/"
    if (fname or subtype) and (not fname or not subtype):
        raise ValueError("Both fname and subtype must be present")
    test_email = email.message.EmailMessage()
    test_email.set_content(content)
    test_email["from"] = "<requestor@mailinator.com>"
    if fname:
        kwargs = {"subtype": subtype, "filename": f"{fname}.{subtype}"}
        if fn_obj == 1:
            if subtype == "csv":
                output = "0"
            else:
                output = b"0"
                kwargs["maintype"] = "application"
        elif fn_obj:
            if subtype != "csv":
                kwargs["maintype"] = "application"
            output = fn_obj
        else:
            if subtype == "csv":
                with open(path + fname, mode="r", encoding="utf-8") as read:
                    output = read.read()
            else:
                with open(path + fname, mode="rb") as bin_read:
                    output = bin_read.read()
                    kwargs["maintype"] = "application"
        test_email.add_attachment(output, **kwargs)
    return test_email


def remove_temps():
    """Clean temporary test data on local disk."""
    for root, dirs, files in os.walk(TEMP_STORAGE_PATH):
        for fn in files:
            if fn != "test_database.db":
                os.remove(os.path.join(root, fn))


def delete_file(fname):
    """Delete specific file from tests temp directory."""
    for root, dirs, files in os.walk(TEMP_STORAGE_PATH):
        for fn in files:
            if fn == fname:
                os.remove(os.path.join(root, fn))


def create_body(r_keys, content):
    """R_keys of tuple added to dict content if not present."""
    if not content:
        content = {}
    for key, value in r_keys:
        if key not in content:
            content[key] = value
    msg_body = ""
    for key, value in content.items():
        msg_body += f"{key}: {value}\n"
    return msg_body


def construct_memory(conf: dict):
    """Construct memory object from configuration."""
    if "tallybot" not in conf:
        raise ValueError(f"Configuration must contain tallybot: {conf}")
    if "database" not in conf["tallybot"]:
        raise ValueError(f"tallybot config must contain database: {conf['tallybot']}")
    db_path = f'sqlite://{conf["tallybot"]["database"]}'
    return membank.LoadMemory(db_path)


def add_memory(setup):
    """Decorate setUp function to include memory."""

    def wrap(self):
        """Add memory method."""
        setup(self)
        self.memory = construct_memory(self.config)

    return wrap


def add_clean_memory(setup):
    """Decorate setUp function to include clean memory."""

    def wrap(self):
        """Clean memory."""
        setup(self)
        self.memory.clean_all_data()
        self.memory = construct_memory(self.config)

    return wrap


def add_unique_name(setup):
    """Decorate setUp function to include unique_name attribute."""

    def wrap(self):
        """Add unique_name to self."""
        self.unique_name = f"{self.__class__.__name__}:{uuid.uuid4()}"
        setup(self)

    return wrap


class AbstractEmailInterfaceTest(TestCase):
    """Common methods for email interface testcases."""

    @add_clean_memory
    @add_memory
    def setUp(self):
        """Enable email interface."""
        super().setUp()
        self.response = email.message.EmailMessage()
        self.email = load_email()
        self.maxDiff = None
        self.email_bot = TestEmailbot()
        self.email_bot.load(self.config)

    def tearDown(self):
        """Clean up of test data."""
        remove_temps()
        self.email_bot.close()

    def do_add_expense(self, content=False):
        """Create expense email and executes."""
        r_keys = [
            ("date", "2022-03-22"),
            ("reference", "expense"),
            ("value", 20),
            ("partner", "Upwork"),
            ("comment", "generic expense"),
        ]
        msg_body = create_body(r_keys, content)
        msg = load_email(content=msg_body, fname="receipt", subtype="pdf")
        msg["subject"] = "do add expense"
        self.do_job_call(msg, assert_status=False)

    def do_add_carwash(self, content=False):
        """Make carwash call."""
        r_keys = [
            ("date", "2022-03-22"),
            ("reference", "carwash"),
            ("value", 20),
        ]
        msg_body = create_body(r_keys, content)
        msg = load_email(content=msg_body, fname="receipt", subtype="pdf")
        msg["subject"] = "do add carwash"
        self.do_job_call(msg)

    def load_transactions(self):
        """Load transactions into memory."""
        self.load_partners()
        self.load_seb_statement()
        msg = load_email(fname="test_upwork_statement", subtype="csv")
        msg["subject"] = "do upwork statement"
        self.do_job_call(msg)

    def load_seb_statement(self):
        """Load seb statement example."""
        msg = load_email(fname="test_seb_statement", subtype="csv")
        msg["subject"] = "do seb statement"
        self.do_job_call(msg)

    def load_upwork_invoices(self):
        """Load upwork invoices."""
        msg = load_email(fname="test_upwork_invoices", subtype="zip")
        msg["subject"] = "do upwork invoices"
        self.do_job_call(msg)

    def load_upwork_statement(self):
        """Load upwork statement."""
        msg = load_email(fname="test_upwork_statement", subtype="csv")
        msg["subject"] = "do upwork statement"
        self.do_job_call(msg)

    def do_job_call(self, msg: email.message.Message = None, assert_status=True):
        """Perform a job call."""
        if not msg:
            msg = self.email
        self.email_bot.ask(msg)
        self.response = self.email_bot.last_message()
        if assert_status:
            self.assertEqual("Done", self.email_bot.last_text(), self.response)

    def get_excel_rows(self):
        """Return openpyxl excel sheet type."""
        self.assertIsNotNone(self.response[1])
        self.assertIsNotNone(self.response[2])
        excel = get_excel(self.response[1])
        sheet = excel.active
        return sheet


def get_excel(binary):
    """Return openpyxl excel object type."""
    exc_file = io.BytesIO(binary)
    return openpyxl.load_workbook(exc_file)


@dataclass
class BrainInterface:
    """Imaginary interface for the brain do_task return."""

    status: str
    attachment: bytes
    attachment_filename: str


def add_date(payload, fmt=None):
    """Add date to dict.

    If there is a year, month or day parameter, construct date out of
    them. Otherwise take today
    """
    today = datetime.date.today()
    dates = ["year", "month", "day"]
    for key in payload:
        if key in dates:
            date = datetime.date(
                year=payload.pop("year", today.year),
                month=payload.pop("month", today.month),
                day=payload.pop("day", today.day),
            )
            payload["date"] = date.strftime(fmt) if fmt else date.isoformat()
            break
    if "date" not in payload:
        payload["date"] = today.strftime(fmt) if fmt else today.isoformat()


class BrainTestCase(TestCase):
    """Abstract testcase for brain interface."""

    maxDiff = None

    @add_unique_name
    @add_clean_memory
    @add_memory
    def setUp(self):
        """Load required methods for brain do_task calls."""
        self.maxDiff = None  # In comparison assertions return full text

    def do(self, cmd, payload=None, attachment=None):
        """Do brain function."""
        return BrainInterface(
            *brain.do_task(
                self.conf,
                self.memory,
                "do_" + cmd,
                payload,
                attachment,
            )
        )

    def do_add_expense(self, **payload):
        """Add new expense."""
        add_date(payload)
        if "value" not in payload:
            payload["value"] = random_amount()
        if "reference" not in payload:
            payload["reference"] = unique_name(self)
        if "comment" not in payload:
            payload["comment"] = "comment" + unique_name()
        response = self.do("add_expense", [payload], b"")
        self.assertEqual("Done", response.status)

    def do_add_partner(self, name):
        """Add partner with name to ledger."""
        response = self.do("create_partner", [{"name": name}])
        self.assertIn("Done", response.status)

    def do_get_ledger(self):
        """Return ledger sheet."""
        response = self.do("get_ledger")
        self.assertEqual("Done", response.status)
        sheet = get_excel(response.attachment).active
        return sheet

    def do_add_bank_statement_line(self, **payload):
        """Import one specific bank statement line."""
        if "debit" in payload:
            payload["value"] = payload.pop("debit")
            payload["side"] = "D"
        if "credit" in payload:
            payload["value"] = payload.pop("credit")
            payload["side"] = "C"
        if "value" not in payload:
            payload["value"] = random_amount()
        if "side" not in payload:
            payload["side"] = "D"
        if "partner" not in payload:
            raise AssertionError("Partner is required")
        add_date(payload, fmt="%d.%m.%Y")
        line = "heading1\nheading2\n"
        line += f";{payload['date']};;{payload['value']};{payload['partner']}"
        line += f";;;;;;;;;;{payload['side']}\n"
        delete_file(f"csv.seb_{datetime.date.today().isoformat()}")
        response = self.do("seb_statement", attachment=line)
        self.assertEqual("Done", response.status)

    def do_transaction(self, **payload):
        """Do direct ledger transaction."""
        add_date(payload)
        if "amount" in payload:
            payload["debit_amount"] = payload.pop("amount")
        if "reference" not in payload:
            payload["reference"] = unique_name(self)
        if "source" not in payload:
            payload["source"] = "/path/to/source/file"
        if "comment" not in payload:
            payload["comment"] = "Comment of the goods"
        if "debit_amount" not in payload:
            payload["debit_amount"] = random_amount()
        response = self.do("transaction", [payload])
        self.assertTrue("Done", response.status)
        return response



"""Tests for brain interface."""
import datetime
import unittest

from tallybot.brain import ledger


class GetIntervalDates(unittest.TestCase):
    """Tests on get_interval_dates function."""

    def setUp(self):
        """Add today."""
        self.today = datetime.date.today()

    def test_quarter(self):
        """Call with filter_by_quarter should return quarter for current year."""
        data = {"filter_by_quarter": "q4"}
        start, end = ledger.get_interval_dates(data)
        must_start = datetime.date(year=self.today.year, month=10, day=1)
        must_end = datetime.date(year=self.today.year + 1, month=1, day=1)
        self.assertEqual(must_start, start)
        self.assertEqual(must_end, end)

    def test_last(self):
        """Calling last will give last full quarter."""
        data = {"filter_by_quarter": "last"}
        start, end = ledger.get_interval_dates(data)
        if self.today.month < 4:
            must_start = datetime.date(year=self.today.year - 1, month=10, day=1)
            must_end = datetime.date(year=self.today.year, month=1, day=1)
        elif self.today.month < 7:
            must_start = datetime.date(year=self.today.year, month=1, day=1)
            must_end = datetime.date(year=self.today.year, month=4, day=1)
        elif self.today.month < 10:
            must_start = datetime.date(year=self.today.year, month=4, day=1)
            must_end = datetime.date(year=self.today.year, month=7, day=1)
        else:
            must_start = datetime.date(year=self.today.year, month=7, day=1)
            must_end = datetime.date(year=self.today.year, month=10, day=1)
        self.assertEqual(must_start, start)
        self.assertEqual(must_end, end)

    def test_january(self):
        """Choose just one month."""
        data = {"filter_by_month": "2022-01"}
        start, end = ledger.get_interval_dates(data)
        must_start = datetime.date(year=2022, month=1, day=1)
        must_end = datetime.date(year=2022, month=2, day=1)
        self.assertEqual(must_start, start)
        self.assertEqual(must_end, end)

    def test_december(self):
        """Choose just one month."""
        data = {"filter_by_month": "2022-12"}
        start, end = ledger.get_interval_dates(data)
        must_start = datetime.date(year=2022, month=12, day=1)
        must_end = datetime.date(year=2023, month=1, day=1)
        self.assertEqual(must_start, start)
        self.assertEqual(must_end, end)



"""Command line utility to start tallybot."""

import argparse
import tomllib
import sys

from zoozl.server import start


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Start tallybot server to listen to different inputs."
    )
    parser.add_argument(
        "config_path",
        type=str,
        help="Path to toml configuration file.",
    )
    args = parser.parse_args()
    with open(args.config_path, "rb") as f:
        config = tomllib.load(f)
    zoozl_cfg = {}
    if "slack" in config:
        if "port" in config["slack"]:
            zoozl_cfg["slack_port"] = config["slack"]["port"]
        else:
            print("No slack port specified in config file. Using default port 8080.")
            zoozl_cfg["slack_port"] = 8080
        if "signing_secret" in config["slack"]:
            zoozl_cfg["slack_signing_secret"] = config["slack"]["signing_secret"]
        else:
            print("No slack signing secret specified in config file. Exiting.")
            sys.exit(1)
        if "workspace_token" in config["slack"]:
            zoozl_cfg["slack_app_token"] = config["slack"]["workspace_token"]
        else:
            print("No slack workspace token specified in config file. Exiting.")
            sys.exit(1)
    start(zoozl_cfg)



"""Holds all memory data structures used by accountant."""

import dataclasses as data
from dataclasses import dataclass
import datetime
import hashlib
import uuid

import membank


DEFAULT_CAR_PLATE = "XX0000"
DEFAULT_CAR_ODO_KM = 0
PARTNER_ACCOUNTS = (
    2310,
    5310,
)  # Accounts that are observed for partner matching


@dataclass
class Conversation:
    """Accountant conversations with people who request actions."""

    talker: str = data.field(default=None, metadata={"key": True})
    ongoing: bool = False
    subject: str = ""
    data: dict = data.field(default_factory=dict)
    attachment: bytes = b""


@dataclass
class ChatSubjectLookup:
    """Accountant subject lookup from chat messages."""

    cmd: str
    sentence: str


class Booking:
    """Automates accounting transactions.

    Decides debit and credit accounts Reassigns correctly closing
    balances also on impacted transactions already stored.
    """

    def __init__(self, text, memory, **kargs):
        """Text value decides what booking will it be."""
        self.memory = memory
        self.comp = self.memory.transaction
        self.load_records()
        text = text.lower()
        splitter = text.find(":")
        set_acc = False
        if splitter != -1:
            text, set_acc = text[0:splitter], int(text[splitter + 1 :])
        if text not in self.records:
            raise RuntimeError(f"No such record '{text}'")
        if set_acc:
            acc_pair = self.records[text]
            if text == "expense":
                acc_pair.insert(0, set_acc)
        kargs["debit"] = self.records[text][0]
        kargs["credit"] = self.records[text][1]
        self.transaction = Transaction(**kargs)
        self.matched_deals = []

    def load_records(self):
        """Add records how to make bookings."""
        self.records = {
            "vat": (5310, 2670),
            "inc_invoice": (7170, 5310),
            "service fee": (5310, 2670),
            "out_invoice": (2310, 6110),
            "hourly": (2670, 2310),
            "fixed price": (2670, 2310),
            "withdrawal fee": (7170, 2670),
            "withdrawal": (2680, 2670),
            "seb-commission": (7640, 5310),
            "upwork-commission": (7170, 2670),
            "seb-expense": (5310, 2620),
            "seb-income": (2620, 2310),
            "seb-upwork-income": (2620, 2680),
            "private-income": (2310, 8900),
            "private-expense": (8900, 5310),
            "fx-profit": (2670, 8150),
            "fx-loss": (8250, 2670),
            "expense": [5310],
            "init-asset": (1230, 3110),
            "asset-depreciate": (7420, 1290),
        }

    def correct_balances(self, node):
        """Apply closing balance recalculations."""
        if not node.right:
            return
        next_node = self.memory.get.node(transaction=node.right)
        next_node.closing = round(node.closing + next_node.value, 2)
        self.memory.put(next_node)
        self.correct_balances(next_node)

    def match_deals(self):
        """Check if other leg of deal is present, if so matches those."""
        to_check = self.transaction
        partner_match = [2310, 5310]
        last_year = datetime.date(year=to_check.date.year - 1, month=12, day=31)
        next_year = datetime.date(year=to_check.date.year + 1, month=1, day=1)
        args = [
            self.comp.partner == to_check.partner,
            self.comp.date > last_year,
            self.comp.date < next_year,
        ]
        if to_check.credit in partner_match:
            args.append(self.comp.debit == to_check.credit)
            args.append(self.comp.debit_currency == to_check.credit_currency)
            args.append(self.comp.debit_stack > 0)
            orphan_deals = self.memory.get(*args)
            self.make_match(orphan_deals, side="debit")
        elif to_check.debit in partner_match:
            args.append(self.comp.credit == to_check.debit)
            args.append(self.comp.credit_currency == to_check.debit_currency)
            args.append(self.comp.credit_stack > 0)
            orphan_deals = self.memory.get(*args)
            self.make_match(orphan_deals, side="credit")

    def make_match(self, orphan_deals, side="debit"):
        """Do a match for a to_check transaction."""
        check = self.transaction
        orphan_deals.sort(key=lambda x: x.date)
        matched_deals = []
        for deal in orphan_deals:
            if side == "debit":
                deal_stack = "debit_stack"
                check_stack = "credit_stack"
            else:
                deal_stack = "credit_stack"
                check_stack = "debit_stack"
            deal_side = getattr(deal, deal_stack)
            check_side = getattr(check, check_stack)
            if deal_side >= check_side:
                setattr(deal, deal_stack, round(deal_side - check_side, 2))
                setattr(check, check_stack, 0)
                self.memory.put(deal)
                matched_deals.append(deal)
                break
            setattr(check, check_stack, round(check_side - deal_side, 2))
            setattr(deal, deal_stack, 0)
            self.memory.put(deal)
            matched_deals.append(deal)
        self.matched_deals = matched_deals

    def convert_currency(self):
        """Handle Upwork currency exchange."""
        item = self.transaction
        mem = self.memory
        if item.debit_currency != item.credit_currency:
            invoices_not_paid = mem.get(
                mem.transaction.credit == 6110,
                mem.transaction.credit_currency == "USD",
                mem.transaction.credit_stack > 0,
            )
            self.make_match(invoices_not_paid, side="credit")
            t_eur = [
                (i.credit_amount - i.credit_stack) / i.rate
                for i in self.matched_deals
            ]
            fx_margin = sum(t_eur) - item.deal_value
            fx_trans = {"date": item.date}
            fx_trans["source"] = item.source
            fx_booking = dict(fx_trans)
            fx_booking["reference"] = item.date.isoformat() + "fx_margin"
            fx_booking["comment"] = item.comment
            fx_booking["debit_amount"] = round(abs(fx_margin), 2)
            if fx_margin > 0:
                fx_type = "fx-profit"
            else:
                fx_type = "fx-loss"
            fx_booking = Booking(fx_type, mem, **fx_booking)
            fx_booking.save()
            fx_commission = dict(fx_trans)
            fx_commission["debit_amount"] = round(
                item.deal_value - item.credit_amount, 2
            )
            fx_commission["reference"] = item.reference + "fx_commission"
            fx_commission["comment"] = "upwork fx commission"
            fx_commission = Booking("upwork-commission", mem, **fx_commission)
            fx_commission.save()

    def save(self):
        """Save transaction into memory."""
        to_save = self.transaction
        booking = self.memory.get(
            "transaction",
            date=to_save.date,
            reference=to_save.reference,
            debit_amount=to_save.debit_amount,
            debit=to_save.debit,
            credit=to_save.credit,
        )
        if not booking:
            self.convert_currency()
            self.match_deals()
            comp = self.memory.transaction
            sides = (
                (
                    comp.debit == to_save.debit,
                    to_save.debit,
                    to_save.debit_amount,
                ),
                (
                    comp.credit == to_save.credit,
                    to_save.credit,
                    -to_save.debit_amount,
                ),
            )
            for side in sides:
                node = Node(
                    transaction=to_save.id,
                    account=side[1],
                    value=side[2],
                    closing=side[2],
                )
                prev_booking = self.memory.get(
                    comp.date <= to_save.date, side[0]
                )
                prev_booking = sorted(prev_booking, key=lambda x: x.date)
                if prev_booking:
                    prev_node = self.memory.get.node(
                        transaction=prev_booking[0].id
                    )
                    node.closing += prev_node.closing
                    node.left = prev_node.transaction
                    if prev_node.right and prev_node.rigth != node.transaction:
                        node.right = prev_node.right
                        self.correct_balances(node)
                        prev_node.right = node.transaction
                        self.memory.put(prev_node)
                self.memory.put(node)
            self.memory.put(to_save)


@dataclass
class Trip:
    """Trips."""

    date: datetime.date
    trip: str
    description: str
    distance: int
    year: int = 0
    month: int = 0
    day: int = 0
    receipt_ref: str = ""
    receipt_litres: float = 0
    receipt_value: float = 0

    def __post_init__(self):
        """Add year, month, day to index."""
        if isinstance(self.distance, str):
            self.distance = int(self.distance)
        if isinstance(self.date, str):
            self.date = datetime.datetime.strptime(self.date, "%Y-%m-%d")
        if isinstance(self.receipt_litres, str):
            self.receipt_litres = float(self.receipt_litres)
        if isinstance(self.receipt_value, str):
            self.receipt_value = float(self.receipt_value)
        self.year = self.date.year
        self.month = self.date.month
        self.day = self.date.day


@dataclass
class Car:
    """Car information."""

    id: str = data.field(default=DEFAULT_CAR_PLATE, metadata={"key": True})
    total_km: int = DEFAULT_CAR_ODO_KM


@dataclass
class PartnerIndex:
    """Names matching patners."""

    text: str
    partner_id: str


@dataclass
class MySelf:
    """Meta storage for the user of accountant."""

    key: str = data.field(metadata={"key": True})
    text: str = ""
    value: float = 0
    dct: dict = data.field(default_factory=dict)
    lst: list = data.field(default_factory=list)
    blob: bytes = b""


@dataclass
class Me:
    """My data."""

    name: str = ""
    address_id: str = ""
    phone: str = ""
    email: str = ""
    bank_name: str = ""
    bank_account: str = ""
    bank_code: str = ""
    rate: float = 0.0
    currency: str = ""


@dataclass
class Partner:
    """Client or Supplier info."""

    name: str
    reg_no: str = ""
    address_id: str = ""
    bank_account: str = ""
    bank_name: str = ""
    invoice_prefix: str = ""
    invoice_row_format: str = ""
    terms: str = ""
    currency: str = ""
    id: str = data.field(default=None, metadata={"key": True})

    def __post_init__(self):
        """Add unique id to partner."""
        if not self.id:
            self.id = str(uuid.uuid1())


@dataclass
class Address:
    """Physical address."""

    address_lines: list = data.field(default_factory=list)
    city: str = ""
    country: str = ""
    postal_code: str = ""
    id: str = data.field(default=None, metadata={"key": True})

    def __post_init__(self):
        """Add unique id if not given."""
        if not self.id:
            self.id = str(uuid.uuid1())


@dataclass
class Node:
    """Node that keeps balances on accounts."""

    transaction: str
    account: int
    value: float
    closing: float
    left: str = ""
    right: str = ""


@dataclass
class Account:
    """Accounts used in General Ledger."""

    id: int
    name: str


@dataclass
class Currency:
    """Currencies and their rates with EUR."""

    date: datetime.date
    currency: str
    rate: float


@dataclass
class Sequences:
    """Generic sequences."""

    relation: str = data.field(metadata={"key": True})
    sequence: int


@dataclass
class TransactionAttachment:
    """Attachments to transactions in GL."""

    transaction_id: str = data.field(metadata={"key": True})
    file_name: str
    file_type: str
    blob: bytes


@dataclass
class Transaction:
    """Transactions in General Ledger."""

    date: datetime.date
    reference: str
    source: str
    comment: str
    debit_amount: float
    credit_amount: float = 0
    partner: str = ""
    debit: int = 0
    credit: int = 0
    deal_value: float = 0
    rate: float = 1
    debit_currency: str = "EUR"
    credit_currency: str = "EUR"
    debit_stack: float = -1
    credit_stack: float = -1
    id: str = data.field(default=None, metadata={"key": True})

    def __post_init__(self):
        """Add unique id to transaction."""
        if not self.id:
            new_id = hashlib.sha1()
            byte_string = self.date.strftime("%Y%m%d").encode()
            byte_string += self.reference.encode()
            byte_string += str(self.debit_amount).encode()
            byte_string += str(self.debit_currency).encode()
            byte_string += str(self.credit_currency).encode()
            byte_string += str(self.debit).encode()
            byte_string += str(self.credit).encode()
            new_id.update(byte_string)
            self.id = new_id.hexdigest()
        if not self.deal_value:
            self.deal_value = round(self.debit_amount / self.rate, 2)
        if not self.credit_amount:
            self.credit_amount = self.debit_amount
        if self.debit_stack == -1:
            self.debit_stack = self.debit_amount
        if self.credit_stack == -1:
            self.credit_stack = self.credit_amount


report_table = {
    "ledger": [
        "date",
        "reference",
        "partner",
        "comment",
        "debit",
        "debit_currency",
        "debit_amount",
        "credit",
        "credit_currency",
        "credit_amount",
        "deal_value",
        "id",
    ],
    "outstanding": [
        "date",
        "reference",
        "partner",
        "debit",
        "debit_currency",
        "debit_amount",
        "debit_stack",
        "credit",
        "credit_currency",
        "credit_amount",
        "credit_stack",
        "deal_value",
    ],
    "partners": [
        "name",
        "reg_no",
        "address_id",
        "bank_account",
        "bank_name",
        "id",
    ],
    "partnerindex": [
        "text",
        "partner_id",
    ],
    "tripsummary": [
        "date",
        "receipt_ref",
        "receipt_litres",
        "trip",
        "description",
        "distance",
    ],
    "": [
        "",
    ],
}


@dataclass
class ReportStruct:
    """Report structure that supports Excel generation."""

    title: str
    items: list
    memory: data.InitVar[membank.LoadMemory]
    report_args: dict = data.field(default_factory=dict)
    callbacks: dict = data.field(default_factory=dict)
    attrs: list = data.field(default_factory=list)
    start: int = 1
    header: list = data.field(default_factory=list)

    def __post_init__(self, memory):
        """Initialise attributes according to title."""
        if self.title not in report_table:
            raise RuntimeError(f"Report '{self.title}' is not supported")
        if not self.attrs:
            self.attrs = report_table[self.title]
        if self.title in ["ledger", "outstanding"]:
            self.callbacks = {
                "date": lambda x: x.strftime("%Y-%m-%d"),
                "partner": lambda x: memory.get.partner(id=x).name if x else "",
            }
        if self.title == "tripsummary":
            self.callbacks = {
                "receipt_litres": lambda x: x if x else "",
                "date": lambda x: x.strftime("%Y-%m-%d"),
            }
            self.start = 10
            end = len(self.items)
            today = datetime.date.today().strftime("%Y-%m-%d")
            self.header = [
                ("D4", "Business Trip Report: {DEFAULT_CAR_PLATE}"),
                ("A3", f"Date: {today}"),
                ("A5", "Fuel: diesel"),
                ("A6", "Consumption: 8l/100km"),
                ("E9", "Total:"),
                ("F9", f"=SUM(F{self.start + 1}:F{self.start + 1 + end})"),
            ]
            if self.report_args and "tahometer" in self.report_args:
                self.header.append(
                    ("A7", f"Tahometer: {self.report_args['tahometer']}")
                )


@dataclass
class TripMetrics:
    """Metrics to perform expense calculation."""

    distance: int = 0
    litres: float = 0



"""Interface on any logical tasks accountant can perform.

>>> with Perform(conf, mem, "do_get_help", {}, b'') as job:
>>>     print(job.status, job.attachment, job.attachment_filename)
>>> assert job.status == "Done"
"""

import csv
from dataclasses import asdict
import datetime
import sys
import traceback
import urllib.request

from tallybot import handlers
from tallybot import memories
from tallybot import learner
from tallybot import managers
from tallybot.lookups import get_mydata, get_partner

from . import frontal_lobe
from .reports import add_reports
from .ledger import add_ledger
from .invoicing import add_invoicing


def do_task(conf, mem, cmd, data=None, attachment=None):
    """Perform a single task in the brain.

    Returns always a tuple of (msg, attachment, attachment_filename).
    msg should contain some text, attachments are optional (could be
    None)
    """
    try:
        with Perform(conf, mem, cmd, data, attachment) as job:
            return (job.status, job.attachment, job.attachment_filename)
    except RuntimeWarning as err:
        return (str(err), None, None)
    except Exception:
        msg = traceback.format_exception(*sys.exc_info())
        msg = "".join(msg)
        msg = "-" * 80 + "\n" + msg + "\n" + "-" * 80
        msg = (
            "Oh my, madness in my brain :sweat: Send this to my developer:\n"
            + msg
        )
        return (msg, None, None)


def generate_help():
    """Generate help message from cmds available in this module."""
    msg = "I did not understand your message\n\n"
    msg += "Send new one with following choices in subject:\n"
    for i in generate_cmd_list():
        msg += " ".join(i.split("_"))
        msg += " - "
        msg += getattr(Perform, i).__doc__
        msg += "\n"
    return msg


def generate_cmd_list():
    """Generate available command list."""
    return [i for i in dir(Perform) if i.startswith("do_")]


def check_required(content, fields):
    """Check if required fields exist in content."""
    misf = []
    for i in fields:
        if i not in content or not content[i]:
            misf.append(i)
    if misf:
        msg = f"Missing required field '{misf[0]}'"
        for i in misf[1:]:
            msg += f", '{i}'"
        raise RuntimeWarning(msg)


def get_invoice_prefix(name):
    """Get invoice prefix from name."""
    return name[:3].upper()


# pylint: disable=too-many-public-methods
@add_reports
@add_ledger
@add_invoicing
class Perform(frontal_lobe.Functions):
    """Map all possible actions to what received msg may ask and execute action
    found."""

    def do_get_help(self):
        """Get help as list of commands available."""
        self.status = ""
        for i in generate_cmd_list():
            self.status += " ".join(i.split("_")) + "\n"

    def do_add_default_partner(self):
        """Add default partner by it's name."""
        data = self.data[0]
        check_required(data, {"name"})
        value = data["name"]
        partner = self.memory.get.partner(name=value)
        if not partner:
            self.status = f"Partner with name '{value}' does not exist"
        else:
            me = memories.MySelf(key="default_partner", text=partner.id)
            self.memory.put(me)

    def do_update_ledger(self):
        """Update transactions given in excel ledger."""
        ws_sheet = handlers.get_excel(self.binary).active
        headers = False
        self.status = ""
        for row in ws_sheet:
            if not headers:
                headers = [i.value.lower() for i in row]
                continue
            book = dict(zip(headers, [i.value for i in row]))
            update = self.memory.get.transaction(id=book.pop("id"))
            for attr, value in book.items():
                if attr == "date":
                    value = datetime.date.fromisoformat(value)
                elif attr == "partner":
                    partner = self.memory.get.partner(name=value)
                    if not partner:
                        self.status += (
                            f"Fail {update.id} as partner {value} not found\n"
                        )
                    value = partner.id
                setattr(update, attr, value)
            self.memory.put(update)
        self.status += "Done"

    def do_private_income(self):
        """Add private income entry.

        date: YYYY-MM-DD
        partner: {partner_name}
        value: {amount}
        """
        data = self.data[0]
        date = datetime.date.fromisoformat(data["date"])
        partner = frontal_lobe.get_partner(self.memory, data.pop("partner"))
        booking = memories.Booking(
            "private-income",
            self.memory,
            **{
                "date": date,
                "reference": "",
                "comment": "private income",
                "partner": partner,
                "source": "",
                "debit_amount": float(data["value"]),
            },
        )
        booking.save()

    def do_delete_transaction(self):
        """Delete a transaction.

        id: str
        """
        del_id = self.data[0]["id"]
        trans = self.memory.get.transaction(id=del_id)
        if not trans:
            self.status = (
                "Transaction with id '{del_id}' not found so not deleted"
            )
        self.memory.delete(trans)

    def do_add_initial_asset(self):
        """Add initial asset with depreciation calculations.

        date: YYYY-MM-DD
        name: str
        reference: str
        value: float
        utility: int (months)
        """
        asset = self.data[0]
        asset["date"] = datetime.datetime.strptime(asset["date"], "%Y-%m-%d")
        path = self.generate_path("inc_invoice", "pdf")
        asset_name = asset.pop("name")
        asset["debit_amount"] = float(asset.pop("value"))
        asset["comment"] = f"Initial asset {asset_name}"
        asset["source"] = path
        utility = int(asset.pop("utility"))
        booking = memories.Booking("init-asset", self.memory, **asset)
        booking.save()
        data = {
            "date": asset["date"],
            "comment": f"Monthly depreciation for {asset_name}",
        }
        data["debit_amount"] = round(asset["debit_amount"] / utility, 2)
        data["source"] = path
        for i in range(utility):
            data["reference"] = (
                f"Depreciation costs on asset '{asset_name}' for month no: {i}"
            )
            if data["date"].month == 12:
                data["date"] = datetime.date(
                    year=data["date"].year + 1,
                    month=1,
                    day=data["date"].day,
                )
            else:
                data["date"] = datetime.date(
                    year=data["date"].year,
                    month=data["date"].month + 1,
                    day=data["date"].day,
                )
            depreciate = memories.Booking(
                "asset-depreciate", self.memory, **data
            )
            depreciate.save()
        handlers.save_file(path, self.binary)

    def do_get_social_report(self):
        """Get report for social tax.

        quarter: YYYY-Qn
        """
        quarter = self.data[0]["quarter"]
        date = datetime.datetime.strptime(quarter[:4], "%Y")
        months = {
            "1": [1, 2, 3],
            "2": [4, 5, 6],
            "3": [7, 8, 9],
            "4": [10, 11, 12],
        }
        self.status = "Social report (profit with 500 deducted already)\n\n"
        for month in months[quarter[-1]]:
            start_date = datetime.date(year=date.year, month=month, day=1)
            end_date = datetime.date(year=date.year, month=month + 1, day=1)
            book = self.memory.get(
                self.memory.transaction.date >= start_date,
                self.memory.transaction.date < end_date,
            )
            income = 0
            expense = 0
            for i in book:
                if 6000 <= i.credit < 7000:
                    income += i.deal_value
                elif 7000 <= i.debit < 8000:
                    expense += i.deal_value
                elif i.debit == 8250:
                    expense += i.deal_value
                elif i.credit == 8150:
                    income += i.deal_value
            profit = round(income - expense - 500, 2)  # profit above 500
            self.status += start_date.strftime("%Y-%m") + f": {profit}\n"

    def do_add_carwash(self):
        """Add expense on carwash.

        date: YYYY-MM-DD
        reference: str
        value: float
        """
        expense = self.data[0]
        expense["partner"] = "Circle K"
        expense["comment"] = "auto car wash"
        expense["expense_account"] = "7760"
        path = self.generate_path("inc_invoice", "pdf")
        expense["path"] = path
        self.make_expense_booking(expense)
        handlers.save_file(path, self.binary)

    def do_add_trip(self):
        """Add new trip.

        date: YYYY-MM-DD
        ----------------
        # Fuel Receipt
        receipt_ref: str
        receipt_litres: float
        receipt_value: float

        attach receipt as pdf
        ----------------
        # Trip
        trip: str
        description: str
        distance: float
        """
        trips = self.data
        if len(trips) > 1:
            raise RuntimeWarning("Only one trip per request")
        trip = trips[0]
        if "receipt_ref" in trip:
            check_required(trip, ["receipt_litres", "receipt_value"])
            trip["trip"] = "Mārupe - Circle K"
            trip["description"] = "add fuel"
            trip["distance"] = 5
            path = self.generate_path("inc_invoice", "pdf")
        check_required(trip, ["trip", "description", "date", "distance"])
        try:
            to_save = memories.Trip(**trip)
        except TypeError as error:
            raise RuntimeWarning(str(error)[11:]) from None
        self.memory.put(to_save)
        car = self.memory.get.car(id=memories.DEFAULT_CAR_PLATE)
        if not car:
            car = memories.Car()
        car.total_km += to_save.distance
        self.memory.put(car)
        if "receipt_ref" in trip:
            handlers.save_file(path, self.binary)

    def do_trip_summary(self):
        """Perform trip summary.

        date: YYYY-MM
        """
        body = self.data
        if len(body) < 1:
            raise RuntimeWarning("There is no date found in message body")
        date = datetime.datetime.strptime(body[0]["date"], "%Y-%m")
        self.prepare_trip_report(date)

    def do_remove_transaction(self):
        """Remove a transaction by it's id."""
        booking = self.data[0]
        item = self.memory.get.transaction(id=booking["id"])
        if item:
            self.memory.delete(item)
        else:
            self.status = f"Id: '{booking['id']}' not found"

    def do_add_fuel_expense(self):
        """Add fuel expense transaction.

        date: YYYY-MM
        tahometer: int
        force:
        """
        expense = self.data[0]
        date = datetime.datetime.strptime(expense["date"], "%Y-%m")
        today = datetime.date.today()
        if (
            date.month != today.month + 1 or today.day > 10
        ) and "force" not in expense:
            msg = "It is recommended to do fuel expense in the first 5 days"
            msg += " of next month or use 'force' to ignore this warning"
            raise RuntimeWarning(msg)
        self.prepare_trip_report(date, tahometer=expense["tahometer"])
        metrics = memories.TripMetrics()
        for i in self.trips:
            metrics.distance += i.distance
            metrics.litres += i.receipt_litres
        if metrics.litres == 0:
            msg = f"There are no fuel receipts for {date.strftime('%Y-%m')}"
            raise RuntimeWarning(msg)
        split = metrics.distance / 100 * 8 / metrics.litres * 100
        for i in self.trips:
            if i.receipt_ref:
                expense = {"split": split, "partner": "Circle K"}
                expense["value"] = i.receipt_value
                expense["date"] = i.date.isoformat()
                expense["reference"] = i.receipt_ref
                expense["comment"] = i.description
                expense["expense_account"] = "7760"
                expense["path"] = self.generate_path("inc_invoice", "pdf")
                self.make_expense_booking(expense)

    def do_private_expense(self):
        """Add private expense.

        date: YYYY-MM-DD
        partner: {partner_name}
        value: {amount}
        """
        expense = self.data[0]
        date = datetime.date.fromisoformat(expense["date"])
        partner = frontal_lobe.get_partner(self.memory, expense.pop("partner"))
        booking = memories.Booking(
            "private-expense",
            self.memory,
            **{
                "date": date,
                "reference": "",
                "comment": "private expense",
                "partner": partner,
                "source": "",
                "debit_amount": float(expense["value"]),
            },
        )
        booking.save()

    def do_add_expense(self):
        """Add expense transaction.

        date: YYYY-MM-DD
        reference: {invoice_no}
        comment: {comment}
        partner: {partner_name}
        value: {amount}
        expense_account: {expense_account} = 7120
        currency: {currency_code} = EUR
        split: {split_percent} = 100
        """
        data = self.data
        l_data = len(data)
        if l_data != 1:
            raise RuntimeWarning(
                f"Exactly one expense invoice per request got {l_data}"
            )
        expense = data[0]
        expense["path"] = self.generate_path("inc_invoice", "pdf")
        self.make_expense_booking(expense)
        handlers.save_file(expense["path"], self.binary)

    def do_get_carinfo(self):
        """Provide car info."""
        car = self.memory.get.car(id=memories.DEFAULT_CAR_PLATE)
        self.status = f"Car {memories.DEFAULT_CAR_PLATE}\n"
        self.status += 10 * "-" + "\n\n"
        self.status += f"Tahometer: {car.total_km}"

    def do_list_partners(self):
        """Provide excel list on existing partners."""
        r_structs = []
        items = self.memory.get("partner")
        r_structs.append(memories.ReportStruct("partners", items, self.memory))
        items = self.memory.get("partnerindex")
        r_structs.append(
            memories.ReportStruct("partnerindex", items, self.memory)
        )
        self.attachment = handlers.ExcelGenerator(*r_structs).binary()
        self.attachment_filename = "partners.xlsx"

    def do_get_myself(self):
        """Provide information about myself."""
        myself = get_mydata(self.memory)
        self.status = ""
        for i, j in asdict(myself).items():
            self.status += i + ": " + str(j) + "\n"

    def do_update_myself(self):
        """Update myself."""
        params = self.data[0]
        myself = get_mydata(self.memory)
        address = managers.AddressManager(self.memory).update(
            params, myself.address_id
        )
        if address:
            params["address_id"] = address.id
        for field in memories.data.fields(memories.Me):
            if field.name in params:
                payload = {"key": field.name}
                if field.type == str:
                    payload["text"] = params[field.name]
                if field.type == float:
                    payload["value"] = float(params[field.name])
                if field.type == list:
                    payload["lst"] = params[field.name]
                if field.type == dict:
                    payload["dct"] = params[field.name]
                if len(payload) == 1:
                    raise ValueError(f"Field {field.name} is not supported")
                ms = memories.MySelf(**payload)
                self.memory.put(ms)

    def do_get_partner(self):
        """Retrieve partner information."""
        partner_name = self.data[0].pop("name")
        partner_id = get_partner(self.memory, partner_name)
        partner = self.memory.get.partner(id=partner_id)
        if not partner:
            raise RuntimeWarning(
                f"Partner with name '{partner_name}' does not exist"
            )
        self.status = ""
        for i, j in asdict(partner).items():
            self.status += i + ": " + str(j) + "\n"
        partner_index = self.memory.get(
            self.memory.partnerindex.partner_id == partner.id
        )
        self.status += "Indexes: "
        for index in partner_index:
            self.status += index.text + ", "

    def do_update_partner(self):
        """Update partner."""
        mgr = managers.PartnerManager(self.memory)
        try:
            mgr.update(self.data[0])
        except AssertionError as e:
            self.status = str(e)

    def do_create_partner(self):
        """Add or update partner in register.

        Use:
            name: {your partner name}
            other_names: {OPTIONAL: similar, similar2}
        """
        for item in self.data:
            if "id" not in item:
                partner = self.memory.get.partner(name=item["name"])
                if partner:
                    self.status += (
                        f"'{partner.name}' exists with id: {partner.id}\n"
                    )
                    continue
            other_names = []
            if "other_names" in item:
                others = [i.strip() for i in item.pop("other_names").split(",")]
                other_names += others
            try:
                item["invoice_prefix"] = get_invoice_prefix(item["name"])
                partner = memories.Partner(**item)
                self.memory.put(partner)
                other_names.append(partner.name)
                for text in other_names:
                    index = self.memory.get.partnerindex(text=text)
                    if index:
                        self.memory.delete(index)
                    self.memory.put(memories.PartnerIndex(text, partner.id))
            except TypeError as error:
                msg = str(error).split(")")[1].strip().capitalize()
                raise RuntimeWarning(msg) from None

    # pylint: disable=R0914, R0912
    # too many statements and local variables
    def do_seb_statement(self):
        """Do seb statement upload."""
        if isinstance(self.binary, str):
            self.binary = self.binary.encode()
        csv_input = self.binary.decode().splitlines()
        csv_reader = csv.reader(csv_input, delimiter=";")
        fname = "seb_" + datetime.date.today().isoformat()
        path = self.generate_path("statement", fname, "csv")
        header = 2
        bookings = []
        errs = []
        for i in csv_reader:
            book_type = False
            if header:
                header -= 1
                continue
            date = datetime.datetime.strptime(i[1], "%d.%m.%Y")
            date = datetime.date(date.year, date.month, date.day)
            value = float(i[3])
            partner = i[4].split("\\")[0].strip()
            if i[14] == "D":
                book_type = "seb-expense"
            elif i[14] == "C":
                if "upwork" in i[4].lower():
                    book_type = "seb-upwork-income"
                else:
                    book_type = "seb-income"
            if "pamatsumma" in i[9] and "KT08093" in i[9]:
                partner = "SEB banka"
            try:
                partner = frontal_lobe.get_partner(self.memory, partner)
            except RuntimeWarning as error:
                errs.append(str(error))
            booking = {
                "date": date,
                "reference": i[10],
                "source": path,
                "comment": i[9],
                "debit_amount": value,
                "partner": partner,
                "book_type": book_type,
            }
            bookings.append(booking)
            # Special auto-bookings
            bookings += learner.get_bookings(booking, i)
        if errs:
            raise RuntimeWarning("\n".join(errs))
        for booking in bookings:
            book_type = booking.pop("book_type")
            booking = memories.Booking(book_type, self.memory, **booking)
            booking.save()
        handlers.save_file(path, self.binary)

    def do_upwork_statement(self):
        """Do upwork statement upload."""
        if isinstance(self.binary, str):
            self.binary = self.binary.encode()
        csv_input = self.binary.decode().splitlines()
        csv_reader = csv.reader(csv_input, delimiter=",")
        fname = "upwork_" + datetime.date.today().isoformat()
        path = self.generate_path("statement", fname, "csv")
        header = True
        for i in csv_reader:
            if header:
                header = False
                continue
            date = datetime.datetime.strptime(i[0], "%b %d, %Y")
            date = datetime.date(date.year, date.month, date.day)
            value = abs(float(i[9]))
            currency = self.exchange_rate(date, "USD")
            partner = ""
            if "VAT" in i[2] or "Service Fee" in i[2] or "Withdrawal" in i[2]:
                partner = frontal_lobe.get_partner(self.memory, "Upwork")
            elif i[6]:
                partner = frontal_lobe.get_partner(self.memory, i[6])
            if i[11]:
                credit_curr = i[11]
                credit_value = float(i[10])
            else:
                credit_curr = "USD"
                credit_value = value
            booking = memories.Booking(
                i[2],
                self.memory,
                **{
                    "date": date,
                    "reference": i[1],
                    "source": path,
                    "comment": i[3],
                    "debit_amount": value,
                    "credit_amount": credit_value,
                    "rate": currency.rate,
                    "debit_currency": "USD",
                    "credit_currency": credit_curr,
                    "partner": partner,
                },
            )
            booking.save()
        handlers.save_file(path, self.binary)

    def do_upwork_invoices(self):
        """Upload invoices from Upwork."""
        zip_file = handlers.get_zip(self.binary)
        for i in zip_file:
            pdf_text = handlers.get_pdf(i)[0]
            invoice = handlers.get_invoice(pdf_text)
            rate = self.exchange_rate(invoice[0], "USD")
            if "Upwork" in invoice[4]:
                inv_type = "inc_invoice"
                partner = frontal_lobe.get_partner(self.memory, "Upwork")
            else:
                inv_type = "out_invoice"
                partner = frontal_lobe.get_partner(self.memory, invoice[5])
            path = self.generate_path(inv_type, "pdf")
            booking = memories.Booking(
                inv_type,
                self.memory,
                **{
                    "date": invoice[0],
                    "reference": invoice[1],
                    "source": path,
                    "comment": invoice[2],
                    "debit_amount": float(invoice[3]),
                    "rate": rate.rate,
                    "debit_currency": "USD",
                    "credit_currency": "USD",
                    "partner": partner,
                },
            )
            booking.save()
            handlers.save_file(path, binary=i)

    def do_accounts_update(self):
        """Upload accounts from external source."""
        url = "http://www.vid.lv/kontu-pl257ns.html"
        with urllib.request.urlopen(url) as response:
            html = response.read()
            parser = frontal_lobe.VIDPageParser(self.memory)
            parser.feed(html.decode(encoding="utf-8"))
